import logging

from nes.instructions import INSTRUCTION_SET, PyNamedInstruction, AddressModes

#from nes import LOG_CPU

class MOS6502:
    """
    Software emulator for MOS Technologies 6502 CPU

    References:
       [1] https://www.masswerk.at/6502/6502_instruction_set.html
       [2] http://www.obelisk.me.uk/6502/addressing.html
       [3] https://www.csh.rit.edu/~moffitt/docs/6502.html

       [4] V-flag:  http://www.6502.org/tutorials/vflag.html
       [5] Decimal mode: http://www.6502.org/tutorials/decimal_mode.html
       [6] PC: http://6502.org/tutorials/6502opcodes.html#PC
       [7] http://forum.6502.org/viewtopic.php?t=1708

       [8] Assembler online: https://www.masswerk.at/6502/assembler.html
       [9] CPU startup sequence:  https://www.pagetable.com/?p=410
       [10] B-flag: http://wiki.nesdev.com/w/index.php/Status_flags#The_B_flag
       [11] V-flag: http://www.righto.com/2012/12/the-6502-overflow-flag-explained.html
       [12] JMP indirect 'bug':  http://forum.6502.org/viewtopic.php?f=2&t=6170

       [13] Undocumented opcodes: http://nesdev.com/undocumented_opcodes.txt
       [14] Undocumented opcodes: http://www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes

       [15] Amazing timing data: http://www.6502.org/users/andre/petindex/local/64doc.txt

       [16] Undocumented opcodes: http://hitmen.c02.at/files/docs/c64/NoMoreSecrets-NMOS6510UnintendedOpcodes-20162412.pdf

    """

    # Masks for the bits in the status register
    SR_N_MASK = 0b10000000  # negative
    SR_V_MASK = 0b01000000  # overflow
    SR_X_MASK = 0b00100000  # unused, but should be set to 1
    SR_B_MASK = 0b00010000  # the "break" flag (indicates BRK was executed, only set on the stack copy)
    SR_D_MASK = 0b00001000  # decimal
    SR_I_MASK = 0b00000100  # interrupt disable
    SR_Z_MASK = 0b00000010  # zero
    SR_C_MASK = 0b00000001  # carry

    # some useful memory locations
    STACK_PAGE = 0x0100           # stack is held on page 1, from 0x0100 to 0x01FF
    IRQ_BRK_VECTOR_ADDR = 0xFFFE  # start of 16 bit location containing the address of the IRQ/BRK interrupt handler
    RESET_VECTOR_ADDR = 0xFFFC    # start of 16 bit location containing the address of the RESET handler
    NMI_VECTOR_ADDR = 0xFFFA      # start of 16 bit location of address of the NMI (non maskable interrupt) handler

    # 6502 is little endian (least significant byte first in 16bit words)
    LO_BYTE = 0
    HI_BYTE = 1

    # cycles taken to do the NMI or IRQ interrupt - (this is a guess, based on BRK, couldn't find a ref for this!)
    INTERRUPT_REQUEST_CYCLES = 7

    OAM_DMA_CPU_CYCLES = 513

    def __init__(self, memory, support_BCD=False, undocumented_support_level=1, aax_sets_flags=False, stack_underflow_causes_exception=True):

        # memory is user-supplied object with read and write methods, allowing for memory mappers, bank switching, etc.
        self.memory = memory

        # create a map from bytecodes to functions for the instruction set
        self.instructions = self._make_bytecode_dict()

        # User-accessible registers
        self.A = None  # accumulator
        self.X = None  # X register
        self.Y = None  # Y register

        # Indirectly accessible registers
        self.PC = None     # program counter (not valid until a reset() has occurred)
        self.SP = None     # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)

        # Flags
        # (stored as an 8-bit register in the CPU)
        # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
        self.N = None
        self.V = None
        self.D = None
        self.I = None
        self.Z = None
        self.C = None

        # CPU cycles since the processor was started or reset
        self.cycles_since_reset = None
        self.started = False

        # debug
        self._previous_PC = 0

        # control behaviour of the cpu
        self.support_BCD = support_BCD
        self.aax_sets_flags = aax_sets_flags
        self.undocumented_support_level = undocumented_support_level
        self.stack_underflow_causes_exception = stack_underflow_causes_exception


    def reset(self):
        """
        Resets the CPU
        """
        # read the program counter from the RESET_VECTOR_ADDR
        self.PC = self._read_word(self.RESET_VECTOR_ADDR)

        # clear the registers
        self.A = 0  # accumulator
        self.X = 0  # X register
        self.Y = 0  # Y register

        # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)
        # reset does three stack pops, so this is set to 0xFD after reset [9]
        self.SP = 0xFD

        # Flags
        # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
        self.N = False
        self.V = False
        self.D = False
        self.I = True   # CPU should "come up with interrupt disable bit set" [7]
        self.Z = False
        self.C = False

        # CPU cycles since the processor was started or reset
        # reset takes 7 cycles [9]
        self.cycles_since_reset = 7


    def _make_bytecode_dict(self):
        """
        Translates the instruction sets into a bytecode->instr dictionary and links the instruction opcodes
        to the functions in this object that execute them (this is done by instruction name).
        """
        instructions = [None] * 256
        for _, instr_set in INSTRUCTION_SET.items():
            for mode, instr in instr_set.modes.items():
                if type(instr.bytecode) == list:
                    # some of the undocumented opcodes have multiple aliases, so in that case a list of
                    # opcodes is supplied
                    for bytecode in instr.bytecode:
                        instructions[bytecode] = PyNamedInstruction(name=instr_set.name,
                                                             bytecode=bytecode,
                                                             mode=mode,
                                                             size_bytes=instr.size_bytes,
                                                             cycles=instr.cycles,
                                                             function=getattr(self, "_{}".format(instr_set.name)),
                                                             )

                else:
                    instructions[instr.bytecode] = PyNamedInstruction(name=instr_set.name,
                                                               bytecode=instr.bytecode,
                                                               mode=mode,
                                                               size_bytes=instr.size_bytes,
                                                               cycles=instr.cycles,
                                                               function=getattr(self, "_{}".format(instr_set.name)),
                                                               )
        return instructions

    def print_status(self):
        """
        Prints a human-readable summary of the CPU status
        """
        print("A: ${0:02x}     X: ${0:02x}     Y: ${0:02x}".format(self.A, self.X, self.Y))
        print("SP: ${:x}".format(self.SP))
        print("STACK: (head) ${0:02x}".format(self.memory.read(self.STACK_PAGE + self.SP)))
        if 0xFF - self.SP > 0:
            print("              ${:x}".format(self.memory.read(self.STACK_PAGE + self.SP + 1)))
        if 0xFF - self.SP > 1:
            print("              ${:x}".format(self.memory.read(self.STACK_PAGE + self.SP + 2)))
        print("Flags:  NV-BDIZC      as byte:  ${:x}".format(self._status_to_byte()))
        print("        {0:08b}".format(self._status_to_byte()))

        print()
        cur_bytecode = self.memory.read(self._previous_PC)
        cur_instr = self.instructions[cur_bytecode]
        cur_data = self.memory.read_block(self._previous_PC + 1, bytes=cur_instr.size_bytes - 1)
        print("last PC: ${:x}    Instr @ last PC: ${:x} - ".format(self._previous_PC, cur_bytecode), end="")
        print(self.format_instruction(cur_instr, cur_data))

        bytecode = self.memory.read(self.PC)
        instr = self.instructions[bytecode]
        data = self.memory.read_block(self.PC + 1, bytes=instr.size_bytes - 1)
        print("next PC: ${:x}    Instr @ next PC: ${:x} - ".format(self.PC, bytecode), end="")
        print(self.format_instruction(instr, data))
        print("cycles: {}".format(self.cycles_since_reset))

    def format_instruction(self, instr, data, caps=True):
        line = ""
        name = instr.name if not caps else instr.name.upper()
        line += '{} '.format(name)
        if instr.mode == AddressModes.IMMEDIATE:
            line += '#${0:02X}'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE:
            line += '${0:02X}'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE_X:
            line += '${0:02X},X'.format(data[0])
        elif instr.mode == AddressModes.ZEROPAGE_Y:
            line += '${0:02X},Y'.format(data[0])
        elif instr.mode == AddressModes.ABSOLUTE:
            line += '${0:04X}'.format(self._from_le(data))
        elif instr.mode == AddressModes.ABSOLUTE_X:
            line += '${0:04X},X'.format(self._from_le(data))
        elif instr.mode == AddressModes.ABSOLUTE_Y:
            line += '${0:04X},Y'.format(self._from_le(data))
        elif instr.mode == AddressModes.INDIRECT_X:
            line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
        elif instr.mode == AddressModes.INDIRECT_Y:
            line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
        elif instr.mode == AddressModes.INDIRECT:
            line += '(${0:04X})'.format(self._from_le(data))
        elif instr.mode == AddressModes.IMPLIED:
            line += ''
        elif instr.mode == AddressModes.ACCUMULATOR:
            line += 'A'
        elif instr.mode == AddressModes.RELATIVE:
            line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
        return line

    def log_line(self):
        """
        Generates a log line in the format of nestest
        """
        str = "{0:04X}  ".format(self.PC)
        bytecode = self.memory.read(self.PC)

        instr = self.instructions[bytecode]
        data = self.memory.read_block(self.PC + 1, bytes=instr.size_bytes - 1)

        str += "{0:02X} ".format(bytecode)
        str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "
        str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "
        str += self.format_instruction(instr, data)

        while len(str) < 48:
            str += " "

        str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:---,--- CYC:{:d}".format(self.A,
                                                                                           self.X,
                                                                                           self.Y,
                                                                                           self._status_to_byte(),
                                                                                           self.SP,
                                                                                           self.cycles_since_reset
                                                                                           )
        return str

    def oam_dma_pause(self):
        cycles = self.OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2
        self.cycles_since_reset += cycles
        return cycles

    def set_reset_vector(self, reset_vector):
        """
        Sets the reset vector (at fixed mem address), which tells the program counter where to start on a reset
        :param reset_vector: 16-bit address to which PC will be set after a reset
        """
        self.memory.write(self.RESET_VECTOR_ADDR, reset_vector & 0x00FF)  # low byte
        self.memory.write(self.RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte

    @staticmethod
    def _from_le(data):
        """
        Create an integer from a little-endian two byte array
        :param data: two-byte little endian array
        :return: an integer in the range 0-65535
        """
        return (data[MOS6502.HI_BYTE] << 8) + data[MOS6502.LO_BYTE]

    def _read_word(self, addr, wrap_at_page=False):
        """
        Read a word at an address and return it as an integer
        If wrap_at_page is True then if the address is 0xppFF, then the word will be read from 0xppFF and 0xpp00
        """
        if wrap_at_page and (addr & 0xFF) == 0xFF:
            # will wrap at page boundary
            byte_lo = self.memory.read(addr)
            byte_hi = self.memory.read(addr & 0xFF00)  # read the second (hi) byte from the start of the page
        else:
            byte_lo = self.memory.read(addr)
            byte_hi = self.memory.read(addr + 1)
        return byte_lo + (byte_hi << 8)

    def trigger_nmi(self):
        """
        Trigger a non maskable interrupt (NMI)
        """
        self._interrupt(self.NMI_VECTOR_ADDR)
        self.cycles_since_reset += self.INTERRUPT_REQUEST_CYCLES
        return self.INTERRUPT_REQUEST_CYCLES

    def trigger_irq(self):
        """
        Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
        """
        if not self.I:
            self._interrupt(self.IRQ_BRK_VECTOR_ADDR)
            self.cycles_since_reset += self.INTERRUPT_REQUEST_CYCLES
            return self.INTERRUPT_REQUEST_CYCLES
        else:
            return 0  # ignored!

    def run_next_instr(self):
        """
        Decode and run the next instruction at the program counter, updating the number of processor
        cycles that have elapsed.  Instructions are taken as atomic, since the 6502 will complete them
        before it responds to interrupts.
        """
        # if the CPU has not yet been started, reset it, which will correctly initialize the PC.
        if not self.started:
            self.started=True
            self.reset()

        # count any extra cycles we incur due to page overruns and branches
        extra_cycles = 0

        # read instruction at current PC
        bytecode = self.memory.read(self.PC)

        # translate to instruction (+ data if needed)
        instr = self.instructions[bytecode]
        #data = self.memory.read_block(self.PC + 1, bytes=instr.size_bytes - 1)
        data = bytearray(2)
        if instr.size_bytes >= 2:
            data[0] = self.memory.read(self.PC + 1)
        if instr.size_bytes == 3:
            data[1] = self.memory.read(self.PC + 2)


        # upon retrieving the opcode, the 6502 immediately increments PC by the opcode size:
        self._previous_PC = self.PC
        self.PC += instr.size_bytes

        # create the right argument based on the addressing mode
        # don't need to know yet if the argument is an address or a literal - the instruction determines this
        immediate = False
        if instr.mode == AddressModes.IMPLIED:
            # all implied instructions are self-contained 1-byte instructions
            arg = None
        elif instr.mode == AddressModes.ACCUMULATOR:
            # the instructions with accumulator targets usually have a memory address as a target, if
            # None is supplied, they target the accumulator
            arg = None
        elif instr.mode == AddressModes.RELATIVE:
            # all relative instructions are control flow instructions that take a relative address
            arg = data[0]
        elif instr.mode == AddressModes.IMMEDIATE:
            arg = data[0]
            immediate = True
        else:
            # memory loading modes:
            address = None
            if instr.mode == AddressModes.ZEROPAGE:
                address = data[0]
            elif instr.mode == AddressModes.ZEROPAGE_X:
                address = (data[0] + self.X) & 0xFF
            elif instr.mode == AddressModes.ZEROPAGE_Y:
                address = (data[0] + self.Y) & 0xFF
            elif instr.mode == AddressModes.ABSOLUTE:
                address = self._from_le(data)
            elif instr.mode == AddressModes.ABSOLUTE_X:
                address = (self._from_le(data) + self.X) & 0xFFFF
                if data[self.LO_BYTE] + self.X > 0xFF and instr.cycles > int(instr.cycles):
                    # extra cycles if cross the page boundary
                    extra_cycles += 1
            elif instr.mode == AddressModes.ABSOLUTE_Y:
                address = (self._from_le(data) + self.Y) & 0xFFFF
                if data[self.LO_BYTE] + self.Y > 0xFF and instr.cycles > int(instr.cycles):
                    # extra cycles if cross the page boundary
                    extra_cycles += 1
            elif instr.mode == AddressModes.INDIRECT:
                # only used for jmp instructions
                # has the jump indirect bug [12] which means it cannot cross page boundaries and instead
                # wraps around, e.g. read from 0x12ff reads 0x12ff and 0x1200
                #address =
                address = self._read_word(self._from_le(data), wrap_at_page=True)
                #address = self._from_le(self.memory.read_block(self._from_le(data), bytes=2))
            elif instr.mode == AddressModes.INDIRECT_X:
                address = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
            elif instr.mode == AddressModes.INDIRECT_Y:
                m = self._read_word(data[0], wrap_at_page=True)
                address = (m + self.Y) & 0xFFFF
                if (m & 0xFF) + self.Y > 0xFF and instr.cycles > int(instr.cycles):
                    # extra cycles if cross the page boundary
                    extra_cycles += 1

            arg = address

        add_cycles = instr.function(arg, immediate)
        extra_cycles += add_cycles if add_cycles else 0

        # update cycle count
        self.cycles_since_reset += int(instr.cycles) + extra_cycles

        # logging
        #logging.log(LOG_CPU, self.log_line(), extra={"source": "CPU"})

        return int(instr.cycles) + extra_cycles

    def _status_to_byte(self, b_flag=0):
        """
        Puts the status register into an 8-bit value.  Bit 6 is set high always, bit 5 (the "B flag") is set according
        to the b_flag argument.  This should be high if called from an instruction (PHP or BRK) instruction, low if
        the call (to push the status to the stack, which is the only use of this instruction within the CPU) came from
        a hardware interrupt (IRQ or NMI).
        """
        return (  self.N * self.SR_N_MASK
                + self.V * self.SR_V_MASK
                + self.SR_X_MASK              # the unused bit should always be set high
                + b_flag * self.SR_B_MASK     # the "B flag" should be set high if called from PHP or BRK [10] low o/w
                + self.D * self.SR_D_MASK
                + self.I * self.SR_I_MASK
                + self.Z * self.SR_Z_MASK
                + self.C * self.SR_C_MASK) & 0xFF   # this final and forces it to be treated as unsigned

    def _status_from_byte(self, sr_byte):
        """
        Sets the processor status from an 8-bit value as found on the stack.
        Bit 5 (B flag) is NEVER set in the status register (but is on the stack), so it is ignored here
        """
        self.N = (sr_byte & self.SR_N_MASK) > 0
        self.V = (sr_byte & self.SR_V_MASK) > 0
        self.D = (sr_byte & self.SR_D_MASK) > 0
        self.I = (sr_byte & self.SR_I_MASK) > 0
        self.Z = (sr_byte & self.SR_Z_MASK) > 0
        self.C = (sr_byte & self.SR_C_MASK) > 0

    def push_stack(self, v):
        """
        Push a byte value onto the stack
        """
        self.memory.write(self.STACK_PAGE + self.SP, v)
        #if self.SP == 0:
        #    if self.stack_overflow_causes_exception:
        #        raise OverflowError("Stack overflow")
        #    else:
        #        self.SP = 0xFF
        self.SP = (self.SP - 1) & 0xFF

    def pop_stack(self):
        """
        Pop (aka 'pull' in 6502 parlance) a byte from the stack
        """
        if self.SP == 0xFF and self.stack_underflow_causes_exception:
            raise OverflowError("Stack underflow")
        self.SP = (self.SP + 1) & 0xFF
        v = self.memory.read(self.STACK_PAGE + self.SP)
        return v

    @staticmethod
    def _neg(v):
        """
        Is the value of v negative in 2s complement (i.e. is bit 7 high)?
        :return: true if v is negative (true if bit 7 of v is 1)
        """
        return (v & 0b10000000) > 0

    @staticmethod
    def _from_2sc(v):
        """
        Convert a 2's complement number to a signed integer
        """
        #neg = (v & 0b10000000) > 0
        #return (v & 0b01111111) if not neg else (v & 0b01111111) - 128
        return (v & 0b01111111) - (v & 0b10000000)

    @staticmethod
    def _to_bcd(v):
        """
        Convert v from an integer (in range 0-100) to (8 bit) binary coded decimal (BCD) representation
        """
        h = int(v / 10) % 10
        l = v - h * 10
        return (h << 4) + l

    @staticmethod
    def _from_bcd(v):
        """
        Convert from 8 bit BCD to integer
        """
        tens = (v & 0xF0) >> 4
        ones = v & 0x0F
        return tens * 10 + ones

    def _set_zn(self, v):
        """
        Sets the Z and N flags from a result value v
        """
        self.Z = (v & 0xFF) == 0  # only care about the bottom 8 bits being zero
        self.N = (v & 0b10000000) > 0

    ##################  The instructions (and their helper functions)  #################################################

    def _adc(self, arg, immediate):
        """
        Add to accumulator with carry.  Result goes into accumulator.
        :param arg: value to be added to accumulator (cannot be None)
        """
        v = self.memory.read(arg) if not immediate else arg

        if self.D and self.support_BCD:
            # On the NES, the D flag has no effect - set support_BCD to False
            # the following horror is based on the behaviour described in [5], Appendix A
            result_bin = (self.A + v + self.C) & 0xFF
            al = (self.A & 0x0F) + (v & 0x0F) + self.C
            if al >= 0x0A:
                al = ((al + 0x06) & 0x0F) + 0x10
            a1 = (self.A & 0xF0) + (v & 0xF0) + al
            a2 = self._from_2sc(self.A & 0xF0) + self._from_2sc(v & 0xF0) + al

            # the value of the V flag in bcd mode is undocumented on the 6502
            # but, according to [5] the following should be correct:
            self.N = (a2 & 0b10000000) > 0
            self.V = ~(-128 <= a2 <= 127)

            if a1 >= 0xA0:
                a1 += 0x60
            self.C = a1 > 255
            result = a1 & 0xFF
            self.Z = result_bin == 0
        else:
            # binary mode (standard)
            result = self.A + v + self.C
            self.C = result > 255
            self.V = (self._neg(self.A) == self._neg(v)) and (self._neg(self.A) != self._neg(result))
            self._set_zn(result)

        # status
        # setting N and Z flags in the usual way is valid in BCD mode because "the N flag contains the high bit of
        # the result of the instruction" [5]


        # result
        self.A = result & 0xFF

    def _and(self, arg, immediate):
        """
        Bitwise AND of accumulator and arg.  Result goes into accumulator.
        """
        v = self.memory.read(arg) if not immediate else arg
        self.A = self.A & v
        self._set_zn(self.A)

    def _asl(self, addr, _):
        """
        Shift left one bit.
        """
        v = self.A if addr is None else self.memory.read(addr)
        result = v << 1

        # status
        self.C = result > 255    # top bit of original is put into carry flag
        self._set_zn(result)

        res_8bit = result & 0xFF

        if addr is None:
            # target the accumulator
            self.A = res_8bit
        else:
            self.memory.write(addr, res_8bit)

    def _jump_relative(self, condition, offset_2sc):
        """
        Jump by offset_2sc (in 2s complement) if the condition is true
        :return: number of extra cycles incurred by this branch conditional
        """
        extra_cycles = 0
        if condition:
            #print("jump")
            extra_cycles = 1  # taking a branch takes at least 1 cycle
            prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
            self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
            if prev_pc_page != self.PC & 0xFF00:
                # but it takes two cycles if the memory page changes
                extra_cycles = 2
        #else:
        #    print("no jump")
        return extra_cycles

    def _bcc(self, offset, _):
        """
        Branch on carry clear
        """
        return self._jump_relative(not self.C, offset)

    def _bcs(self, offset, _):
        """
        Branch on carry set
        """
        return self._jump_relative(self.C, offset)

    def _beq(self, offset, _):
        """
        Branch if zero flag is set
        """
        return self._jump_relative(self.Z, offset)

    def _bit(self, addr, _):
        """
        Manipulates the status register by setting the N and V flags to those bits of the value in the address
        given, and sets the zero flag if A & v == 0
        """
        v = self.memory.read(addr)
        self.N = (v & self.SR_N_MASK) > 0
        self.V = (v & self.SR_V_MASK) > 0
        self.Z = (self.A & v) == 0

    def _bmi(self, offset, _):
        """
        Branch on result minus (i.e. negative flag N is set)
        """
        return self._jump_relative(self.N, offset)

    def _bne(self, offset, _):
        """
        Branch if zero flag is not set
        """
        return self._jump_relative(not self.Z, offset)

    def _bpl(self, offset, _):
        """
        Branch on result positive (i.e. negative flag N is not set)
        """
        #print(self.N, not self.N)
        return self._jump_relative(not self.N, offset)

    def _interrupt(self, interrupt_vector, is_brk=False):
        """
        Interrupt routine, followed (with variations) by NMI, IRQ and BRK
          1)
        """
        # push PC + 1 to the stack, high bit first
        v = self.PC + (1 if is_brk else 0)
        self.push_stack((v & 0xFF00) >> 8)  # high byte
        self.push_stack(v & 0x00FF)  # low byte
        # push the processor status to the stack
        # BUT note that the B flag ON THE STACK COPY ONLY is now set
        sr = self._status_to_byte(b_flag=True if is_brk else False)
        self.push_stack(sr)
        #addr = self._from_le(self.memory.read_block(interrupt_vector, bytes=2))
        addr = self._read_word(interrupt_vector)
        self.PC = addr
        if not is_brk:
            # if this is not a brk instruction, set the interrupt disable flag
            self.I = True

    def _brk(self, _, __):
        """
        Force break, which simulates an interrupt request.
        BRK, unlike other interrupts (IRQ and NMI), pushes PC + 1 to the stack (high byte first as usual).
        The reason for this may have been to allow brk to be dropped in in place of two byte instructions to allow
        debugging, but it is a quirk of the BRK instruction.  BRK also sets the B flag in the value of SR pushed to
        the stack
        """
        #
        self._interrupt(self.IRQ_BRK_VECTOR_ADDR, is_brk=True)

    def _bvc(self, offset, _):
        """
        Branch on on overflow clear (V == 0)
        """
        return self._jump_relative(not self.V, offset)

    def _bvs(self, offset, _):
        """
        Branch on on overflow set (V == 1)
        """
        return self._jump_relative(self.V, offset)

    def _clc(self, _, __):
        """
        Clear carry flag (set C:=0)
        """
        self.C = False

    def _cld(self, _, __):
        """
        Clear decimal flag (set D:=0)
        """
        self.D = False

    def _cli(self, _, __):
        """
        Clear interrupt disable flag (set I:=0)
        """
        self.I = False

    def _clv(self, _, __):
        """
        Clear value flag (set V:=0)
        """
        self.V = False

    def _compare(self, v0, v):
        self._set_zn(v0 - v)
        self.C = v0 >= v

    def _cmp(self, arg, immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.A, v)

    def _cpx(self, arg, immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.X, v)

    def _cpy(self, arg, immediate):
        """
        Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
        :return:
        """
        v = self.memory.read(arg) if not immediate else arg
        self._compare(self.Y, v)

    def _dec(self, addr, _):
        """
        Decrement memory in addr by 1
        :param addr:
        :param _:
        :return:
        """
        v = (self.memory.read(addr) - 1) & 0xFF
        self._set_zn(v)
        self.memory.write(addr, v)

    def _dex(self, _, __):
        """
        Decrement X by 1
        :param addr:
        :param _:
        :return:
        """
        self.X = (self.X - 1) & 0xFF
        self._set_zn(self.X)

    def _dey(self, _, __):
        """
        Decrement Y by 1
        :param addr:
        :param _:
        :return:
        """
        self.Y = (self.Y - 1) & 0xFF
        self._set_zn(self.Y)

    def _eor(self, arg, immediate):
        """
        XOR A with value and put result back into A
        :param addr:
        :param _:
        :return:
        """
        v = self.memory.read(arg) if not immediate else arg
        self.A = v ^ self.A
        self._set_zn(self.A)

    def _inc(self, addr, _):
        """
        Increment memory in addr by 1
        :param addr:
        :param _:
        :return:
        """
        v = (self.memory.read(addr) + 1) & 0xFF
        self._set_zn(v)
        self.memory.write(addr, v)

    def _inx(self, _, __):
        """
        Increment X by 1
        :param addr:
        :param _:
        :return:
        """
        self.X = (self.X + 1) & 0xFF
        self._set_zn(self.X)

    def _iny(self, _, __):
        """
        Increment Y by 1
        :param addr:
        :param _:
        :return:
        """
        self.Y = (self.Y + 1) & 0xFF
        self._set_zn(self.Y)

    def _jmp(self, addr, _):
        """
        Jump to the (16 bit) address addr.  In the case of jump instructions jmp and jsr,
        "absolute" addressing is more like immediate addressing, in that the jump goes directly
        to the memory address specified rather than reading the value there and using that (which
        is "indirect" addressing in the case of jumps, so this function behaves always like it got
        an immediate address
        """
        self.PC = addr

    def _jsr(self, addr, _):
        """
        Jump to a subroutine at the value in memory[addr].  Addr here is an absolute (16 bit) location.
        In the case of jump instructions jmp and jsr,
        "absolute" addressing is more like immediate addressing, in that the jump goes directly
        to the memory address specified rather than reading the value there and using that (which
        is "indirect" addressing in the case of jumps, so this function behaves always like it got
        an immediate address
        """
        # save PC - 1 to the stack, high byte first
        v = self.PC - 1
        self.push_stack((v & 0xFF00) >> 8)
        self.push_stack(v & 0x00FF)
        self.PC = addr

    def _lda(self, arg, immediate):
        """
        Load A from value / memory
        :return:
        """
        self.A = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.A)

    def _ldx(self, arg, immediate):
        """
        Load X from value / memory
        :return:
        """
        self.X = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.X)

    def _ldy(self, arg, immediate):
        """
        Load Y from value / memory
        :return:
        """
        self.Y = self.memory.read(arg) if not immediate else arg
        self._set_zn(self.Y)

    def _lsr(self, addr, _):
        """
        Logical shift right
        :param addr:
        :param _:
        :return:
        """
        v = self.A if addr is None else self.memory.read(addr)
        result = (v >> 1) & 0xFF

        # status
        self.C = v & 0x01
        self.N = False
        self.Z = result == 0

        if addr is None:
            # target the accumulator
            self.A = result
        else:
            self.memory.write(addr, result)

    def _nop(self, _, __):
        """
        No-op.  This one is easy.
        :return:
        """
        pass

    def _ora(self, arg, immediate):
        """
        Bitwise OR with accumulator; result put into accumulator.
        :return:
        """
        v = self.memory.read(arg) if not immediate else arg
        self.A = self.A | v
        self._set_zn(self.A)

    def _pha(self, _, __):
        """
        Push A onto stack
        :param _:
        :param __:
        :return:
        """
        self.push_stack(self.A)

    def _php(self, _, __):
        """
        Push status register onto stack
        :param _:
        :param __:
        :return:
        """
        v = self._status_to_byte(b_flag=True)  # b-flag should be set True in this case [10]
        self.push_stack(v)

    def _pla(self, _, __):
        """
        Pull A from stack
        :param _:
        :param __:
        :return:
        """
        self.A = self.pop_stack()
        self._set_zn(self.A)

    def _plp(self, _, __):
        """
        Pull processor status from the stack
        :return:
        """
        v = self.pop_stack()
        self._status_from_byte(v)

    def _rol(self, addr, _):
        """
        Rotate one bit left (including carry bit in the rotation, i.e. C rotates into lsb and msb rotates into C)
        :return:
        """
        v = self.A if addr is None else self.memory.read(addr)
        result = (v << 1) + self.C

        # status
        self.C = result > 255
        self._set_zn(result)

        if addr is None:
            # target the accumulator
            self.A = result & 0xFF
        else:
            self.memory.write(addr, result & 0xFF)

    def _ror(self, addr, _):
        """
        Rotate one bit right (including carry bit in the rotation, i.e. C rotates into msb and lsb rotates into C)
        :return:
        """
        v = self.A if addr is None else self.memory.read(addr)
        result = (v >> 1) + (self.C << 7)

        # status
        self.C = v & 0x01
        self._set_zn(result)

        if addr is None:
            # target the accumulator
            self.A = result & 0xFF
        else:
            self.memory.write(addr, result & 0xFF)

    def _rti(self, _, __):
        """
        Return from an interrupt.  Gets the PC of the last execution point and the flag register from the
        stack.  (PC is stored on the stack, not PC-1 as with jsr/rts)
        :return:
        """
        flags = self.pop_stack()
        addr_lo = self.pop_stack()
        addr_hi = self.pop_stack()
        self._status_from_byte(flags)
        self.PC = (addr_hi << 8) + addr_lo

    def _rts(self, _, __):
        """
        Returns from a subroutine (see jsr for the jump subroutine)
        :return:
        """
        # fetch PC-1 from the stack, low byte first
        addr_lo = self.pop_stack()
        addr_hi = self.pop_stack()
        # restore PC to the stack
        self.PC = (addr_hi << 8) + addr_lo + 1

    def _sbc(self, arg, immediate):
        """
        Subtract memory plus carry flag from A and put the result in A (i.e. A := A - M - ~C).
        The carry flag is set if A - M - ~C >= 0, i.e. if no borrow is necessary (allows chaining of
        subtract instructions).
        The overflow (V) flag is set if there is a (signed) overflow.
        Overflow occurs when the result is either too negative or too positive to be held in the available 8 bits.
        This can happen in two cases:
          1) when A is +ve and v is -ve; this should create a +ve number, so if result is -ve that is overflow
          2) when A is -ve and v is +ve; this should create a -ve number, so if result is +ve that is overflow
        """
        v = self.memory.read(arg) if not immediate else arg
        borrow = not self.C   # incoming borrow is indicated by not Carry  (borrow is either 0 or 1)

        result = self.A - v - borrow     # arithmetic is done unsigned
        self.C = result >= 0  # not Carry == borrow; borrow if res<0, so C==not(res<0) == res>=0
        # overflow - see notes in description above
        # I think this is correct and that the borrow doesn't matter here
        self.V = (self._neg(self.A) != self._neg(v)) and (self._neg(v) == self._neg(result))
        self._set_zn(result)

        if self.D and self.support_BCD:
            # according to [5], all the flags behave as if in binary mode on the 6502 even in BCD mode
            al = (self.A & 0x0F) - (v & 0x0F) + self.C - 1
            if al < 0:
                al = ((al - 0x06) & 0x0F) - 0x10
            a = (self.A & 0xF0) - (v & 0xF0) + al
            if a < 0:
                a -= 0x60
            result = a
        else:
            # the NES can have the D bit set and acts as in binary mode (I think)
            pass
            #raise Exception("BCD mode not supported (start CPU with support_BCD=True)")

        self.A = result & 0xFF

    def _sec(self, _, __):
        """
        Set carry flag
        """
        self.C = True

    def _sed(self, _, __):
        """
        Set decimal flag
        """
        self.D = True

    def _sei(self, _, __):
        """
        Set interrupt disable flag
        """
        self.I = True

    def _sta(self, addr, _):
        """
        Store A in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.A)

    def _stx(self, addr, _):
        """
        Store X in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.X)

    def _sty(self, addr, _):
        """
        Store Y in memory.  No flags are set.
        :return:
        """
        self.memory.write(addr, self.Y)

    def _tax(self, _, __):
        """
        Transfer A to X
        :return:
        """
        self.X = self.A
        self._set_zn(self.X)

    def _tay(self, _, __):
        """
        Transfer A to Y
        :return:
        """
        self.Y = self.A
        self._set_zn(self.Y)

    def _tsx(self, _, __):
        """
        Transfer SP to X
        :return:
        """
        self.X = self.SP
        self._set_zn(self.X)

    def _txa(self, _, __):
        """
        Transfer X to A
        :return:
        """
        self.A = self.X
        self._set_zn(self.A)

    def _txs(self, _, __):
        """
        Transfer X to SP
        :return:
        """
        self.SP = self.X

    def _tya(self, _, __):
        """
        Transfer Y to A
        :return:
        """
        self.A = self.Y
        self._set_zn(self.A)

    ################################## The undocumented instructions (see [13] and others) #############################
    # todo: add the undocumented-level feature to the instruction list and remove the per-function tests here

    def _dop(self, _, __):
        """
        Undocumented.
        Double NOP instruction.  Luckily it is easy to implement!
        """
        pass

    def _top(self, _, __):
        """
        Undocumented.
        Triple NOP instruction.
        """
        pass

    def _kil(self, _, __):
        """
        Undocumented.
        Shuts down the processor.  No recovery is possible after this has executed, short of a reset.
        """
        if self.undocumented_support_level >= 1:
            raise ValueError("KIL instruction.  Processor halted.")

    def _lax(self, arg, immediate):
        """
        Undocumented.
        Load A and X from value / memory
        :return:
        """
        if self.undocumented_support_level >= 1:
            v = self.memory.read(arg) if not immediate else arg
            self.A = v
            self.X = v
            self._set_zn(self.A)

    def _aax(self, addr, _):
        """
        Undocumented.
        AND X register with accumulator and store result in memory.
        Status flags: N,Z  (or are they?  nestest does not think so...)
        """
        if self.undocumented_support_level >= 1:
            v = self.A & self.X
            self.memory.write(addr, v)
            if self.aax_sets_flags:
                # the behaviour of this seems unclear, nestest does not think it does this, [13] does
                self._set_zn(v)

    def _dcp(self, addr, _):
        """
        Undocumented.
        This opcode DECs the contents of a memory location and then CMPs the result with the A register.
        Equivalent to DEC oper, CMP oper [14]
        """
        if self.undocumented_support_level >= 1:
            v = (self.memory.read(addr) - 1) & 0xFF
            self.memory.write(addr, v)
            self._compare(self.A, v)

    def _isc(self, addr, _):
        """
        Undocumented.
        Increase memory by one, then subtract memory from accumulator (with borrow). Status flags: N,V,Z,C
        """
        if self.undocumented_support_level >= 1:
            v = (self.memory.read(addr) + 1) & 0xFF
            self.memory.write(addr, v)
            self._sbc(v, immediate=True)

    def _slo(self, addr, _):
        """
        Undocumented.
        Shift left one bit in memory, then OR accumulator with memory.  Status flags: N,Z,C
        Equivalent to ASL, ORA
        """
        if self.undocumented_support_level >= 1:
            self._asl(addr, None)
            self._ora(self.memory.read(addr), immediate=True)

    def _rla(self, addr, _):
        """
        Undocumented.
        RLA ROLs the contents of a memory location and then ANDs the result with the accumulator.
        Equivalent to ROL, AND
        """
        if self.undocumented_support_level >= 1:
            self._rol(addr, None)
            self._and(self.memory.read(addr), immediate=True)

    def _rra(self, addr, _):
        """
        Undocumented.
        RRA RORs the contents of a memory location and then ADCs the result with the accumulator.
        Equivalent to ROR, ADC
        """
        if self.undocumented_support_level >= 1:
            self._ror(addr, None)
            self._adc(self.memory.read(addr), immediate=True)

    def _sre(self, addr, _):
        """
        Undocumented.
        Shift right one bit in memory, then EOR accumulator with memory.
        Equivalent to LSR, EOR
        """
        if self.undocumented_support_level >= 1:
            self._lsr(addr, None)
            self._eor(self.memory.read(addr), immediate=True)

    ##################### The arcane undocumented instructions (see [13] and others) ###################################

    def _arr(self, arg, _):
        """
        UNTESTED
        Undocumented.  Mostly used for piracy (jk).
        ANDs the contents of the A register with an immediate value and then RORs the result.
        Equivalent to AND oper, ROR A
        """
        if self.undocumented_support_level >= 2:
            self._and(arg, immediate=True)
            self._ror(None, None)

    def _asr(self, arg, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A register with an immediate value and then LSRs the result.
        Equivalent to AND oper, LSR A
        """
        if self.undocumented_support_level >= 2:
            self._and(arg, immediate=True)
            self._lsr(None, None)

    def _atx(self, arg, _):
        """
        UNTESTED
        Undocumented.
        ORs the A register with #$EE, ANDs the result with an immediate value, and then stores the result in
        both A and X.
        Equivalent to ORA #$EE, AND oper, TAX
        """
        if self.undocumented_support_level >= 2:
            self._and(arg | 0xEE, immediate=True)
            self.X = self.A

    def _aac(self, arg, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A register with an immediate value and then
        moves bit 7 of A into the Carry flag.  This opcode works basically
        identically to AND #immed. except that the Carry flag is set to the same
        state that the Negative flag is set to.  [14]
        """
        if self.undocumented_support_level >= 2:
            self.A = self.A & arg
            self._set_zn(self.A)
            self.C = (self.A & 0b10000000) > 0

    def _axa(self, addr, _):
        """
        UNTESTED
        Undocumented.
        Sores the result of A AND X AND the high byte of the target address of the operand +1 in memory [14, 15]
        WARNING:  There seem to be multiple conflicting descriptions of this opcode. (see e.g. [13])
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            v = (self.X & self.A) & hp1
            self.memory.write(addr, v)

    def _axs(self, addr, _):
        """
        UNTESTED
        Undocumented.
        AND the contents of the A and X registers (without changing the contents of either register) and
        stores the result in memory. [16]
        """
        if self.undocumented_support_level >= 2:
            self.memory.write(addr, self.A & self.X)

    def _lar(self, addr, _):
        """
        UNTESTED
        Undocumented.
        AND memory with stack pointer, transfer result to accumulator, X register and stack pointer. [16]
        """
        if self.undocumented_support_level >= 2:
            v = self.memory.read(addr) & self.SP
            self.A = v
            self.X = v
            self.SP = v

    def _sxa(self, addr, _):
        """
        UNTESTED
        Undocumented.
        AND X register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            self.memory.write(addr, self.X & hp1)

    def _sya(self, addr, _):
        """
        UNTESTED
        Undocumented.
        AND Y register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
        """
        if self.undocumented_support_level >= 2:
            v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
            self.memory.write(addr, v)
            self._set_zn(v)

    def _xaa(self, arg, _):
        """
        UNTESTED
        Undocumented.
        This opcode ORs the A register with CONST, ANDs the result with X. ANDs the result with an immediate value, and
        then stores the result in A.
        Known to be extremely unstable, including temperature dependent! [16]
        """
        if self.undocumented_support_level >= 2:
            CONST = 0x00
            self.A = (self.A | CONST) & self.X & arg

    def _xas(self, addr, _):
        """
        UNTESTED
        Undocumented.
        ANDs the contents of the A and X registers (without changing the contents
        of either register) and transfers the result to the stack pointer. It then ANDs that result with the
        contents of the high byte of the target address of the operand +1 and stores that final result in
        memory
        """
        if self.undocumented_support_level >= 2:
            hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
            self.SP = self.A & self.X
            self.memory.write(addr, self.SP & hp1)




# cython: profile=True, boundscheck=True, nonecheck=False, language_level=3
#import pyximport; pyximport.install()

import logging

DEF MIRROR_HORIZONTAL = (0, 0, 1, 1)
DEF MIRROR_VERTICAL = (0, 1, 0, 1)
DEF MIRROR_ONE_LOWER = (0, 0, 0, 0)
DEF MIRROR_ONE_UPPER = (1, 1, 1, 1)

from .bitwise cimport bit_high

cdef class NESCart:
    """
    NES Cartridge interface
    """
    def __init__(self):
        pass

    cdef unsigned char read(self, int address):
        """
        Read from cartridge memory attached to the CPU's memory bus at the (CPU) address specified
        """
        raise NotImplementedError()

    cdef void write(self, int address, unsigned char value):
        """
        Write the given byte to cartridge memory attached to the CPU's memory bus at the (CPU) address specified
        """
        raise NotImplementedError()

    cdef unsigned char read_ppu(self, int address):
        """
        Read from cartridge memory attached to the PPU's memory bus (i.e. in the VRAM address space) at the address
        specified
        """
        raise NotImplementedError()

    cdef void write_ppu(self, int address, unsigned char value):
        """
        Write a byte to the cartridge memory attached to the PPU's memory bus (i.e. in the VRAM address space) at the
        address specified
        """
        raise NotImplementedError()

    cdef void irq_tick(self):
        """
        Function called on rising edges of PPU address line A12, which is used in some carts (MMC3 and derivatives) to
        tick an IRQ counter.  Carts without PPU triggered IRQ functionality can ignore this.
        """
        pass


### Mapper 0 (aka NROM) ################################################################################################

cdef class NESCart0(NESCart):
    """
    Basic NES Cartridge (Type 0 / MMC0).  Consists of up to 8kB RAM, 32kB PRG ROM, 8kB CHR ROM
    """
    def __init__(self,
                 prg_rom_data=None,
                 chr_rom_data=None,
                 prg_rom_writeable=False,   # whether or not the program memory is ROM or RAM
                 ram_size_kb=8,
                 nametable_mirror_pattern=(0, 0, 1, 1)
                 ):
        super().__init__()

        # initialize ram (this ram is on the CPU bus)
        if ram_size_kb not in [2, 4, 8]:
            raise ValueError("Cart 0 ram size should be exactly 2, 4 or 8kB")
        self.ram_size = ram_size_kb * 1024

        # initialize prg rom from supplied data  (CPU connected)
        if prg_rom_data:
            # can use a None value to skip this initializing if calling from e.g. other mappers
            self.prg_rom_size = len(prg_rom_data)
            if self.prg_rom_size not in [16 * 1024, 32 * 1024]:
                raise ValueError("Cart 0 program rom size should be exactly 16 or 32kB")
            for i in range(self.prg_rom_size):
                # could use e.g. memcpy, but this syntax is clearer and don't care about speed here
                self.prg_rom[i] = prg_rom_data[i]   # copy prg rom data into the prg ROM

        # initialize chr rom from supplied data (or create empty RAM if no data supplied)
        if chr_rom_data:
            if len(chr_rom_data) != M0_CHR_MEM_SIZE:
                raise ValueError("Cart 0 chr rom size should be exactly 8kB")
            for i in range(M0_CHR_MEM_SIZE):
                self.chr_mem[i] = chr_rom_data[i]
            # ROM data was supplied, so this is a ROM chip and is not writeable
            self.chr_mem_writeable = False
        else:
            # if no chr_rom_data is supplied, chr_mem should be a RAM chip
            self.chr_mem_writeable = True

        self.prg_start_addr = PRG_ROM_START
        for i in range(4):
            self.nametable_mirror_pattern[i] = nametable_mirror_pattern[i]
        self.prg_rom_writeable = prg_rom_writeable

    cdef unsigned char read(self, int address):
        if address < PRG_ROM_START:
            # RAM access
            return self.ram[address % self.ram_size]
        else:
            # program ROM access
            return self.prg_rom[address % self.prg_rom_size]

    cdef void write(self, int address, unsigned char value):
        if address < PRG_ROM_START:
            # RAM access
            self.ram[address % self.ram_size] = value
        elif self.prg_rom_writeable:
            # program ROM access (but only if allowed to be written, otherwise ignored)
            self.prg_rom[address % self.prg_rom_size] = value

    cdef unsigned char read_ppu(self, int address):
        return self.chr_mem[address % M0_CHR_MEM_SIZE]

    cdef void write_ppu(self, int address, unsigned char value):
        if self.chr_mem_writeable:
            self.chr_mem[address % M0_CHR_MEM_SIZE] = value


### Mapper 1 (aka MMC1, SxROM) #########################################################################################

cdef class NESCart1(NESCart):
    """
    NES Cartridge type 1 (MMC1).  Common cartridge with bank-switched prg and chr rom.  Controlled by an internal shift
    register that writes one of four five-bit internal registers (ctrl, prg_bank, chr_bank[2]) depending on the address
    of the fifth write.

    There are several MMC1 variants!  Not sure how to tell these all of these apart from iNES v1 header...
        MMC1A always has PRG RAM enabled
        MMC1B has PRG RAM enabled by default
        MMC1C has PRG RAM disabled by default

    Of the many MMC1 (SxROM) variants [4], a few have unusual properties that require additional modelling:

        - SOROM, SXROM and SZROM have bank switched RAM, and of these, only SOROM was available outside Japan [4].

        - SOROM has 16kb of PRG RAM (2 banks); bank selection is by bit 3 of the chr_bank register
          - this is the only bank-switched prg_ram variant supported at present

        - SXROM only has two Japan only games [5] and so it is not supported

        - SUROM has 512kb PRG ROM
           - there are only two games listed for this config in the cart database [2], and only a couple in the list [3]
             therefore, it is not (yet?) supported

        - SZROM (not implemented) has a different control bit for PRG RAM

    References:
        [1] https://wiki.nesdev.com/w/index.php/MMC1
        [2] http://bootgod.dyndns.org:7777/
        [3] http://tuxnes.sourceforge.net/nesmapper.txt
        [4] https://wiki.nesdev.com/w/index.php/SxROM
        [5] http://bootgod.dyndns.org:7777/search.php?unif_op=LIKE+`%25%40%25`&unif=SXROM
    """
    def __init__(self,
                 prg_rom_data=None,
                 chr_rom_data=None,
                 chr_mem_writeable=False, # if True, the CHR memory is actually RAM, not ROM
                 prg_ram_size_kb=8,       # size of prg RAM; must be in 8kb multiples, can be 0. RAM is at 0x6000-0x7FFF
                 nametable_mirror_pattern=MIRROR_HORIZONTAL,
                ):
        super().__init__()

        # copy the prg rom data to the memory banks
        self.num_prg_banks = int(len(prg_rom_data) / M1_PRG_ROM_BANK_SIZE)
        if self.num_prg_banks * M1_PRG_ROM_BANK_SIZE != len(prg_rom_data):
            raise ValueError("prg_rom_data size ({} bytes) is not an exact multiple of bank size ({} bytes)".format(len(prg_rom_data), M1_PRG_ROM_BANK_SIZE))

        for bnk in range(self.num_prg_banks):
            for i in range(M1_PRG_ROM_BANK_SIZE):
                self.banked_prg_rom[bnk][i] = prg_rom_data[bnk * M1_PRG_ROM_BANK_SIZE + i]

        if self.num_prg_banks > M1_MAX_PRG_BANKS:
            # this is a 512kb prg rom ROM
            raise ValueError("512kb MMC1 ROMs are not currently supported.  Sorry Dragon Warrior III/IV.")

        # copy the chr rom data to the memory banks
        self.num_chr_banks = int(len(chr_rom_data) / M1_CHR_ROM_BANK_SIZE)
        if self.num_chr_banks * M1_CHR_ROM_BANK_SIZE != len(chr_rom_data):
            raise ValueError("chr_rom_data size ({} bytes) is not an exact multiple of bank size ({} bytes)".format(len(chr_rom_data), M1_CHR_ROM_BANK_SIZE))

        for bnk in range(self.num_chr_banks):
            for i in range(M1_CHR_ROM_BANK_SIZE):
                self.banked_chr_rom[bnk][i] = chr_rom_data[bnk * M1_CHR_ROM_BANK_SIZE + i]

        self.chr_mem_writeable = chr_mem_writeable

        # todo is this the right thing to do here?
        if self.num_chr_banks == 0:
            # if there is no chr rom, make an 8kb bank of ram
            self.num_chr_banks = 2
            self.chr_mem_writeable = True

        # set the startup state of the cart
        self.shift_ctr = 0  # shift counter, triggers when it reaches 5
        self.shift = 0      # shift register to accumulate writes to registers
        # there is only one prg bank switch, which can have one of three different effects depending on ctrl register
        self.prg_bank = 0
        self.chr_bank[:] = [0, 0]  # two chr bank registers
        self.prg_ram_bank = 0
        self.ctrl = 0

        self.num_prg_ram_banks = int(prg_ram_size_kb * BYTES_PER_KB / M1_PRG_RAM_BANK_SIZE)
        if self.num_prg_ram_banks * M1_PRG_RAM_BANK_SIZE != prg_ram_size_kb * BYTES_PER_KB:
            raise ValueError("PRG RAM size must be a multiple of 8kb")
        if self.num_prg_ram_banks > 2:
            raise ValueError("PRG RAM with more than two banks (>16kb) is not supported.  This seems to be a rare format (SUROM or SXROM).")

        # set the nametable mirror pattern
        self.nametable_mirror_pattern[:] = nametable_mirror_pattern

        # set the nametable mirror bits (bottom two) of ctrl to the correct value
        if tuple(nametable_mirror_pattern) == MIRROR_ONE_LOWER:
            self.ctrl |= 0
        elif tuple(nametable_mirror_pattern) == MIRROR_ONE_UPPER:
            self.ctrl |= 1
        elif tuple(nametable_mirror_pattern) == MIRROR_VERTICAL:
            self.ctrl |= 2
        elif tuple(nametable_mirror_pattern) == MIRROR_HORIZONTAL:
            self.ctrl |= 3

        # sets the PRG ROM bank mode to 3 (fix last bank at 0xC000, switch 16kb bank at 0x8000)
        # see control register section in [1], which has a note on startup condition of ctrl
        self.ctrl |= 0b00001100

    cdef void write(self, int address, unsigned char value):
        cdef int bank=0

        if address < M1_PRG_RAM_START:
            # shouldn't be writing here
            pass
        elif M1_PRG_RAM_START <= address < M1_CTRL_REG_START:
            if self.prg_bank & 0b10000 == 0:  # bit 4 being low is prg_ram enable
                if self.num_prg_ram_banks > 1 and (self.chr_bank[0] & 0b1000) > 0:
                    # bank-switched prg ram; only SOROM is supported of this type, and this has the selection bit in
                    # bit 3 of the chr_bank register
                    bank = (self.chr_bank[0] >> 3) & 1
                self.ram[bank][address % M1_PRG_RAM_BANK_SIZE] = value
        elif address >= M1_CTRL_REG_START:
            # everything else is a write to the shift register:
            self._write_shift(address, value)

    cdef void _write_shift(self, int address, unsigned char value):
        """
        Writes one bit to the internal shift register, or resets it.  On the fifth (non-reset) write, the value in the
        shift register is written to one of the internal registers; which one is determined by the address of that
        fifth write.
        :param address:  address to write to; used on fifth write to select register to be written
        :param value: value to write, only msb (reset) and lsb (data) bits are used
        """
        cdef unsigned char data, reset

        data = value & 1          # lsb of value is the data
        reset = (value >> 7) & 1  # msb of value is reset bit

        if reset:
            # "Reset shift register and write Control with (Control OR $0C), locking PRG ROM at $C000-$FFFF to the
            # last bank."
            self.shift = 0
            self.shift_ctr = 0
            self.ctrl |= 0x0C   # sets the PRG ROM bank mode to 3 (fix last bank at 0xC000, switch 16kb bank at 0x8000)
            return

        # put the data into the shift register, shifting the register right and putting the data in 4 (i.e. the 5th bit)
        self.shift = (self.shift >> 1) | (data << 4)
        self.shift_ctr += 1

        #print("{:06b} in shift (#{})".format(self.shift, self.shift_ctr))

        if self.shift_ctr == 5:
            # this was the final write to the register and we can now transfer the contents of the shift register to the
            # internal register in question.  We also reset the shift register (to 0) and counter.
            if M1_CTRL_REG_START <= address < M1_CHR_REG_0_START:
                self.ctrl = self.shift
                # if we've written ctrl, we might have to update the mirror pattern:
                self._set_nametable_mirror_pattern()
                #print("{:06b} written to ctrl".format(self.ctrl))
            elif M1_CHR_REG_0_START <= address < M1_CHR_REG_1_START:
                self.chr_bank[0] = self.shift
                #print("{} written to chr_bank[0]".format(self.chr_bank[0]))
            elif M1_CHR_REG_1_START <= address < M1_PRG_REG_START:
                self.chr_bank[1] = self.shift
                #print("{} written to chr_bank[1]".format(self.chr_bank[1]))
            elif M1_PRG_REG_START <= address:
                self.prg_bank = self.shift
                #print("{} written to prg_bank".format(self.prg_bank))
            #else:
                #print("UNMATCHED SHIFT FINISH {:04X}".format(address))

            self.shift_ctr = 0
            self.shift = 0

    cdef void _set_nametable_mirror_pattern(self):
        """
        Updates the nametable mirror pattern to correspond to the value in ctrl
        """
        cdef int mirror_pattern
        mirror_pattern = self.ctrl & 0b11
        if mirror_pattern == 0:
            self.nametable_mirror_pattern[:] = MIRROR_ONE_LOWER
        elif mirror_pattern == 1:
            self.nametable_mirror_pattern[:] = MIRROR_ONE_UPPER
        elif mirror_pattern == 2:
            self.nametable_mirror_pattern[:] = MIRROR_VERTICAL
        elif mirror_pattern == 3:
            self.nametable_mirror_pattern[:] = MIRROR_HORIZONTAL

    cdef unsigned char read(self, int address):
        """
        Read from the PRG memory (connected to CPU).  The address and the internal state of the ctrl and prg_bank
        registers determine the exact location of the memory read.
        :param address: address to read from
        :return: the one byte contents of the memory at address
        """
        cdef int bank=0, mode=-1
        cdef unsigned char value

        if M1_PRG_RAM_START <= address < M1_PRG_ROM_BANK0_START:
            # There might be more than 1 bank of RAM or non at all, need to check it is enabled
            if self.prg_bank & 0b10000 == 0:  # bit 4 being low is prg_ram enable
                # prg-ram enabled
                if self.num_prg_ram_banks > 1 and (self.chr_bank[0] & 0b1000) > 0:
                    # bank-switched prg ram; only SOROM is supported of this type, and this has the selection bit in
                    # bit 3 of the chr_bank register
                    bank = (self.chr_bank[0] >> 3) & 1
                value = self.ram[bank][address % M1_PRG_RAM_BANK_SIZE]
            else:
                # open-bus behaviour
                # todo: should be open bus behaviour, but not sure yet how to implement this
                value = 0
            return value
        elif M1_PRG_ROM_BANK0_START <= address < M1_PRG_ROM_BANK1_START:
            # read from prg rom bank 0 (this is an address between 0x8000 and 0xBFFF)
            # which bank is bank 0 is determined by the prg rom mode (in ctrl) and the prg_bank register
            mode = (self.ctrl & 0b1100) >> 2
            if mode == 0 or mode == 1:   # 32 kb mode
                # in 32kb mode, ignore lower bit (and in this case we are in the lower of the two 16kb banks, so lsb=0)
                bank = (self.prg_bank & 0b1110) % self.num_prg_banks
            elif mode == 2:
                # bank 0 is fixed to first bank at 0x8000 (but that is the area that address is in, so read first bank)
                # (in 512kb ROMs, even this "fixed" bank is switched to the second page)
                bank = 0
            elif mode == 3:
                # bank 0 is switched according to prg_bank
                bank = (self.prg_bank & 0b1111) % self.num_prg_banks
        elif M1_PRG_ROM_BANK1_START <= address:
            # read from prg rom bank 1 (this is an address between 0xC000 and 0xFFFF)
            # which bank is bank 1 is determined by the prg rom mode (in ctrl) and the prg_bank register
            mode = (self.ctrl & 0b1100) >> 2
            if mode == 0 or mode == 1:  # 32 kb mode
                # in 32kb mode, ignore lower bit of the prg_bank, but now we are in the upper of the two banks, so the
                # lsb must be set to 1
                bank = ((self.prg_bank & 0b1110) + 1) % self.num_prg_banks
            elif mode == 2:
                # bank 1 is switched according to prg_bank
                bank = (self.prg_bank & 0b1111) % self.num_prg_banks
            elif mode == 3:
                # bank 1 is fixed to last bank
                bank = self.num_prg_banks - 1

        value = self.banked_prg_rom[bank][address % M1_PRG_ROM_BANK_SIZE]
        return value

    cdef int _get_chr_bank(self, int address):
        """
        Get the bank in the case of a read/write to the chr memory
        :param address:
        :return:
        """
        cdef int mode_8kb, mask
        # which addressing mode?  8kb or (if this is false) 4kb
        mode_8kb = (self.ctrl & 0b10000) == 0
        # is the address in bank 0 or bank 1?
        address_bank = address >= M1_CHR_ROM_BANK1_START
        # just use the lower line for addressing if only have 2 banks; higher lines might be in use for other things
        mask = 0b00001 if self.num_chr_banks <= 2 else 0b11111
        if mode_8kb:
            # use the address in chr_bank[0], ignoring the low bit.
            # in this case bank (chr_bank[0] & 0b11110) is at 0x0000
            #               and (chr_bank[0] & 0b11110) + 1 is at 0x1000
            mask &= 0b11110
            return (self.chr_bank[0] & mask) + address_bank
        else:
            # 4kb mode
            return self.chr_bank[address_bank] & mask

    cdef void write_ppu(self, int address, unsigned char value):
        """
        Write a byte to CHR memory; should only work in carts with RAM here.
        """
        if not self.chr_mem_writeable:
            # this is a ROM, so do nothing
            return
        self.banked_chr_rom[self._get_chr_bank(address)][address % M1_CHR_ROM_BANK_SIZE] = value

    cdef unsigned char read_ppu(self, int address):
        """
        Read from the chr rom; the exact piece of chr rom read depends on the internal state of the chr_bank registers
        :param address: address to read from
        :return: a byte read from chr rom
        """
        return self.banked_chr_rom[self._get_chr_bank(address)][address % M1_CHR_ROM_BANK_SIZE]


### Mapper 2 (aka UNROM, UOROM) ########################################################################################
# ref: https://wiki.nesdev.com/w/index.php/UxROM

cdef class NESCart2(NESCart0):
    """
    NES Cartridge Type 2 (UNROM / UOROM)
    Much like Mapper 0 above, but has switchable banks of 16kB prg_rom
    """
    def __init__(self,
             prg_rom_data=None,
             chr_rom_data=None,
             prg_rom_writeable=False,   # whether or not the program memory is ROM or RAM
             ram_size_kb=8,
             nametable_mirror_pattern=(0, 0, 1, 1),
             emulate_bus_conflicts=False   # whether or not to emulate bus conflicts on prg_bank select writes
             ):
        super().__init__(prg_rom_data=None,    # don't want to write prg_rom_data to the prg_rom because will be too big
                         chr_rom_data=chr_rom_data,
                         prg_rom_writeable=prg_rom_writeable,
                         ram_size_kb=ram_size_kb,
                         nametable_mirror_pattern=nametable_mirror_pattern,
                         )

        self.prg_bank = 0

        # Write the prg_rom_data to the ROM banks 16kB at a time.  Hopefully the ROM size is divisible by 16kB...
        self.num_prg_banks = int(len(prg_rom_data) / M2_PRG_ROM_BANK_SIZE)

        if self.num_prg_banks * M2_PRG_ROM_BANK_SIZE != len(prg_rom_data):
            raise ValueError("prg_rom_data size is not an integer multiple of 16kB")
        if not self.num_prg_banks in [2**i for i in range(5)]:
            raise ValueError("Number of banks should be a power of two <=16.")

        # Copy the data to the banked_prg_rom banks
        for bnk in range(self.num_prg_banks):
            for i in range(16 * 1024):
                self.banked_prg_rom[bnk][i] = prg_rom_data[bnk * 16 * 1024 + i]

        self.emulate_bus_conflicts = emulate_bus_conflicts

    cdef unsigned char read(self, int address):
        if address < PRG_ROM_START:
            # RAM access
            return self.ram[address % self.ram_size]
        elif address < M2_FIXED_PRG_ROM_START:
            # 0x8000 - 0xBFFF is access to banked prg_rom
            return self.banked_prg_rom[self.prg_bank][address % M2_PRG_ROM_BANK_SIZE]
        else:
            # 0xC000 - 0xFFFF is access to the last 16kB bank
            return self.banked_prg_rom[self.num_prg_banks - 1][address % M2_PRG_ROM_BANK_SIZE]

    cdef void write(self, int address, unsigned char value):
        if address < PRG_ROM_START:
            # RAM access
            self.ram[address % self.ram_size] = value
        else:
            # any write to 0x8000 - 0xFFFF is treated as a write to the bank select register
            if self.emulate_bus_conflicts:
                # Bus conflicts occur because in some carts the ROM outputs the value in the ROM at the given address at the
                # same time as the CPU is trying to write a value to this prg_bank register.  If they are not in agreement
                # a bus conflict occurs and undefined behaviour can result.  A lot of programs get around this by writing to
                # a location containing a value that agrees with the value being written (e.g. write 1 to 0x8000 when 0x8000
                # contains 1, etc.).  It seems like in general 0 wins in bus conflicts, so they can be implemented by ANDing
                # the value being written with the contents of the memory at the given address.
                # see: https://wiki.nesdev.com/w/index.php/UxROM
                #      https://wiki.nesdev.com/w/index.php/Bus_conflict
                value &= self.read(address)

            self.prg_bank = value % self.num_prg_banks


### Mapper 4 (aka MMC3) ################################################################################################
# ref: https://wiki.nesdev.com/w/index.php/MMC3

cdef class NESCart4(NESCart):
    def __init__(self,
                 prg_rom_data=None,
                 chr_rom_data=None,
                 chr_mem_writeable=False, # if True, the CHR memory is actually RAM, not ROM
                 prg_ram_size_kb=8,       # size of prg RAM; must 8kb or 0
                 nametable_mirror_pattern=MIRROR_HORIZONTAL,
                 mirror_pattern_fixed=False,  # if true, the mirror pattern cannot be changed dynamically
                 interrupt_listener=None,
                ):
        super().__init__()

        if prg_ram_size_kb != 8 and prg_ram_size_kb != 0:
            raise ValueError("PRG RAM size must be 8kb or 0")

        # copy the prg rom data to the memory banks
        self.num_prg_banks = int(len(prg_rom_data) / M4_PRG_ROM_BANK_SIZE)
        print("PRG bytes: {}".format(len(prg_rom_data)))
        print("Num PRG banks: {}".format(self.num_prg_banks))
        if self.num_prg_banks * M4_PRG_ROM_BANK_SIZE != len(prg_rom_data):
            raise ValueError("prg_rom_data size ({} bytes) is not an exact multiple of bank size ({} bytes)".format(len(prg_rom_data), M4_PRG_ROM_BANK_SIZE))
        if self.num_prg_banks > M4_MAX_PRG_BANKS:
            raise ValueError("Too much prg rom data ({} bytes) for a MMC3 cartridge (max is 512kb)".format(len(prg_rom_data)))

        for bnk in range(self.num_prg_banks):
            for i in range(M4_PRG_ROM_BANK_SIZE):
                self.banked_prg_rom[bnk][i] = prg_rom_data[bnk * M4_PRG_ROM_BANK_SIZE + i]

        # copy the chr rom data to the memory banks
        self.num_chr_banks = int(len(chr_rom_data) / M4_CHR_ROM_BANK_SIZE)
        print("CHR bytes: {}".format(len(chr_rom_data)))
        print("Num CHR banks: {}".format(self.num_chr_banks))
        if self.num_chr_banks * M4_CHR_ROM_BANK_SIZE != len(chr_rom_data):
            raise ValueError("chr_rom_data size ({} bytes) is not an exact multiple of bank size ({} bytes)".format(len(chr_rom_data), M4_CHR_ROM_BANK_SIZE))

        for bnk in range(self.num_chr_banks):
            for i in range(M4_CHR_ROM_BANK_SIZE):
                self.banked_chr_rom[bnk][i] = chr_rom_data[bnk * M4_CHR_ROM_BANK_SIZE + i]

        self.chr_mem_writeable = chr_mem_writeable

        # todo is this the right thing to do here?
        if self.num_chr_banks == 0:
            # if there is no chr rom, make a full-sized block of chr ram (overwrites user requested chr_mem_writeable)
            self.num_chr_banks = 256
            self.chr_mem_writeable = True

        # set the startup state of the cart
        self.interrupt_listener = interrupt_listener
        self.bank_register[:] = [0, 2, 4, 5, 6, 7, 0, 1]   # copied these from somewhere (didn't help with GunNac)
        self.chr_a12_inversion = False
        self.prg_bank_mode = 0
        self.bank_select = 0
        self.prg_ram_enable = True
        self.prg_ram_protect = False
        self.irq_reload_value = 0xFF
        self.irq_enabled = True
        self.irq_counter = 0xFF

        # set the nametable mirror pattern
        self.nametable_mirror_pattern[:] = nametable_mirror_pattern
        self.mirror_pattern_fixed = mirror_pattern_fixed

    cdef unsigned char read(self, int address):
        cdef int window_slot
        cdef int banks[4]

        if M4_PRG_RAM_START <= address < M4_PRG_ROM_START:
            if self.prg_ram_enable:
                return self.ram[address % M4_PRG_RAM_SIZE]
            else:
                # todo: open bus behaviour
                return 0
        elif M4_PRG_ROM_START <= address:
            if self.prg_bank_mode == 0:
                # bank map in mode 0
                banks[0] = self.bank_register[6] & 0b00111111
                banks[1] = self.bank_register[7] & 0b00111111
                banks[2] = self.num_prg_banks - 2
                banks[3] = self.num_prg_banks - 1
            else:
                # bank map in mode 1
                banks[0] = self.num_prg_banks - 2
                banks[1] = self.bank_register[7] & 0b00111111
                banks[2] = self.bank_register[6] & 0b00111111
                banks[3] = self.num_prg_banks - 1

            window_slot = (address >> 13) & 0b11     # address lines A13 and A14 determine the prg rom slot of address
            return self.banked_prg_rom[banks[window_slot] % self.num_prg_banks][address % M4_PRG_ROM_BANK_SIZE]

    cdef void write(self, int address, unsigned char value):
        if M4_PRG_RAM_START <= address < M4_PRG_ROM_START:
            # ram write
            if self.prg_ram_enable and not self.prg_ram_protect:
                self.ram[address % M4_PRG_RAM_SIZE] = value
        elif BANK_REG_START <= address < MIRROR_PROTECT_REG_START:
            if address & 1 == 0:
                # even address => bank select write
                self.chr_a12_inversion = bit_high(value, 7)
                self.prg_bank_mode = bit_high(value, 6)
                self.bank_select = value & 0b111
            else:
                # odd address => bank data write (write one of the bank registers)
                self.bank_register[self.bank_select] = value

        elif MIRROR_PROTECT_REG_START <= address < IRQ_LATCH_RELOAD_REG_START:
            if address & 1 == 0 and not self.mirror_pattern_fixed:
                # sets nametable mirror pattern
                if value & 1:
                    self.nametable_mirror_pattern[:] = MIRROR_HORIZONTAL
                else:
                    self.nametable_mirror_pattern[:] = MIRROR_VERTICAL
            else:
                # set prg_ram access
                self.prg_ram_enable = bit_high(value, 7)
                self.prg_ram_protect = bit_high(value, 6)
        elif IRQ_LATCH_RELOAD_REG_START <= address < IRQ_ACTIVATE_START:
            if address & 1 == 0:
                self.irq_reload_value = value
            else:
                self.irq_reload = True   # reload irq at next tick
        else:
            # irq is disabled by any write to an even address in this range and activated by a write to any odd address
            # in this range, so we can just use the bottom bit of the address line to set the irq_enabled flag
            self.irq_enabled = address & 1
            if not self.irq_enabled:
                # if interrupts are disabled, any pending interrupts are acknowledged;
                # todo: this will also clear pending interrupts from other IRQ producing devices (the APU), but these
                # should not both be in operation at the same time anyway, since then they can be attempting to drive
                # the line in different directions anyway.
                self.interrupt_listener.reset_irq()

    cdef unsigned int _get_ppu_bank(self, int address):
        cdef int window_slot  # the 1kb slot in the window that this address belongs to
        cdef int banks[8]     # the banks that are mapped to each of the eight slots in the chr address space (each 1kb)
        cdef int double_bank_start, single_bank_start  # which slots the 2kb and 1kb banks start in, respectively

        if not self.chr_a12_inversion:
            double_bank_start, single_bank_start = 0, 4
        else:
            double_bank_start, single_bank_start = 4, 0

        # the chr memory banks corresponding to the 1kb window slots
        banks[double_bank_start + 0] = self.bank_register[0] & 0b11111110
        banks[double_bank_start + 1] = (self.bank_register[0] & 0b11111110) + 1
        banks[double_bank_start + 2] = self.bank_register[1] & 0b11111110
        banks[double_bank_start + 3] = (self.bank_register[1] & 0b11111110) + 1
        banks[single_bank_start + 0] = self.bank_register[2]
        banks[single_bank_start + 1] = self.bank_register[3]
        banks[single_bank_start + 2] = self.bank_register[4]
        banks[single_bank_start + 3] = self.bank_register[5]

        window_slot = (address >> 10) & 0b111   # address lines A10, A11 and A12 determine the slot
        return banks[window_slot] % self.num_chr_banks

    cdef unsigned char read_ppu(self, int address):
        cdef unsigned int bank = self._get_ppu_bank(address)
        return self.banked_chr_rom[bank][address % M4_CHR_ROM_BANK_SIZE]

    cdef void write_ppu(self, int address, unsigned char value):
        cdef unsigned int bank
        if self.chr_mem_writeable:
            bank = self._get_ppu_bank(address)
            self.banked_chr_rom[bank][address % M4_CHR_ROM_BANK_SIZE] = value

    cdef void irq_tick(self):
        if self.irq_reload:
            # if a reload has been triggered, do that and reset the reload flag
            # (add 1 to the counter here because this will be immediately decremented below and there should be N + 1
            # periods between IRQs)
            self.irq_counter = self.irq_reload_value + 1
            self.irq_reload = False

        if self.irq_counter > 0:
            self.irq_counter -= 1

        logging.log(logging.INFO, "IRQ counter = {}".format(self.irq_counter), extra={"source": "Cart"})

        # currently, this implements "normal" or "new" behaviour (https://wiki.nesdev.com/w/index.php/MMC3), so that if
        # the reload value is zero, an irq will be triggered every cycle rather than only once, but this can vary
        # between MMC3 cartridges because on some the irq is only triggered as the counter is decremented to zero.
        # For those old-style chips, writing to reload with the counter at zero will trigger another single IRQ
        # todo: support "old" MMC3A behaviour
        if self.irq_counter == 0:
            self.irq_reload = True  # reload counter next time
            if self.irq_enabled:
                self.interrupt_listener.raise_irq()






from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
import Cython.Compiler.Options
Cython.Compiler.Options.annotate = True

extensions = [Extension("cycore.*", ["nes/cycore/*.pyx"])]
extensions = cythonize(extensions, compiler_directives={"language_level": 3, "profile": False, "boundscheck": False, "nonecheck": False, "cdivision": True}, annotate=True)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pyntendo",
    version="0.0.8",
    author="James Murphy",
    author_email="jkmurphy314@gmail.com",
    description="A Nintendo Entertainment System (NES) emulator in Python and Cython",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jameskmurphy/nes",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    #ext_modules = cythonize("nes/cycore/*.pyx")
    ext_modules = extensions
)


# cython: profile=True, boundscheck=False, nonecheck=False

from .memory cimport NESMappedRAM


cdef enum:

    # Masks for the bits in the status register
    SR_N_MASK = 0b10000000  # negative
    SR_V_MASK = 0b01000000  # overflow
    SR_X_MASK = 0b00100000  # unused, but should be set to 1
    SR_B_MASK = 0b00010000  # the "break" flag (indicates BRK was executed, only set on the stack copy)
    SR_D_MASK = 0b00001000  # decimal
    SR_I_MASK = 0b00000100  # interrupt disable
    SR_Z_MASK = 0b00000010  # zero
    SR_C_MASK = 0b00000001  # carry

    # some useful memory locations
    STACK_PAGE = 0x0100           # stack is held on page 1, from 0x0100 to 0x01FF
    IRQ_BRK_VECTOR_ADDR = 0xFFFE  # start of 16 bit location containing the address of the IRQ/BRK interrupt handler
    RESET_VECTOR_ADDR = 0xFFFC    # start of 16 bit location containing the address of the RESET handler
    NMI_VECTOR_ADDR = 0xFFFA      # start of 16 bit location of address of the NMI (non maskable interrupt) handler

    # 6502 is little endian (least significant byte first in 16bit words)
    LO_BYTE = 0
    HI_BYTE = 1

    # cycles taken to do the NMI or IRQ interrupt - (this is a guess, based on BRK, couldn't find a ref for this!)
    INTERRUPT_REQUEST_CYCLES = 7
    OAM_DMA_CPU_CYCLES = 513

    # Used to represent None for integer arguments in some places
    ARG_NONE = -1


cdef class MOS6502:
    cdef NESMappedRAM memory        # the main memory; important that type is decalared so can use c-calling
    cdef object instructions        # the instructions in the original python format; still used for logging etc.

    cdef unsigned char A, X, Y      # registers
    cdef int PC, SP                 # program and stack pointers
    cdef int N, V, D, I, Z, C       # status bits

    cdef long long cycles_since_reset     # cycles since the processor was reset
    cdef int aax_sets_flags, undocumented_support_level, stack_underflow_causes_exception

    # instruction size data table
    cdef int instr_size_bytes[256]


    cdef void make_instruction_data_tables(self)
    cdef int run_instr(self, unsigned char opcode, unsigned char data[2])
    cdef int dma_pause(self, int pause_type, int count)
    cdef int _from_le(self, unsigned char* data)
    cdef int _read_word(self, int addr, int wrap_at_page)
    cdef int trigger_nmi(self)
    cdef int trigger_irq(self)
    cdef int run_next_instr(self)
    cdef unsigned char _status_to_byte(self, bint b_flag)
    cdef void _status_from_byte(self, unsigned char sr_byte)
    cdef void push_stack(self, unsigned char v)
    cdef unsigned char pop_stack(self)
    cdef int _neg(self, int v)
    cdef int _from_2sc(self, unsigned char v)
    cdef void _set_zn(self, int v)

    ##################  The instructions (and their helper functions)  #################################################
    # only the documented instructions are here for now

    cdef int _adc(self, int arg, int immediate)
    cdef int _and(self, int arg, int immediate)
    cdef int _asl(self, int addr, int _)
    cdef int _jump_relative(self, int condition, int offset_2sc)
    cdef int _bcc(self, int offset, int _)
    cdef int _bcs(self, int offset, int _)
    cdef int _beq(self, int offset, int _)
    cdef int _bit(self, int addr, int _)
    cdef int _bmi(self, int offset, int _)
    cdef int _bne(self, int offset, int _)
    cdef int _bpl(self, int offset, int _)
    cdef _interrupt(self, int interrupt_vector, int is_brk)
    cdef int _brk(self, int _, int __)
    cdef int _bvc(self, int offset, int _)
    cdef int _bvs(self, int offset, int _)
    cdef int _clc(self, int _, int __)
    cdef int _cld(self, int _, int __)
    cdef int _cli(self, int _, int __)
    cdef int _clv(self, int _, int __)
    cdef void _compare(self, int v0, int v)
    cdef int _cmp(self, int arg, int immediate)
    cdef int _cpx(self, int arg, int immediate)
    cdef int _cpy(self, int arg, int immediate)
    cdef int _dec(self, int addr, int _)
    cdef int _dex(self, int _, int __)
    cdef int _dey(self, int _, int __)
    cdef int _eor(self, int arg, int immediate)
    cdef int _inc(self, int addr, int _)
    cdef int _inx(self, int _, int __)
    cdef int _iny(self, int _, int __)
    cdef int _jmp(self, int addr, int _)
    cdef int _jsr(self, int addr, int _)
    cdef int _lda(self, int arg, int immediate)
    cdef int _ldx(self, int arg, int immediate)
    cdef int _ldy(self, int arg, int immediate)
    cdef int _lsr(self, int addr, int _)
    cdef int _nop(self, int _, int __)
    cdef int _ora(self, int arg, int immediate)
    cdef int _pha(self, int _, int __)
    cdef int _php(self, int _, int __)
    cdef int _pla(self, int _, int __)
    cdef int _plp(self, int _, int __)
    cdef int _rol(self, int addr, int _)
    cdef int _ror(self, int addr, int _)
    cdef int _rti(self, int _, int __)
    cdef int _rts(self, int _, int __)
    cdef int _sbc(self, int arg, int immediate)
    cdef int _sec(self, int _, int __)
    cdef int _sed(self, int _, int __)
    cdef int _sei(self, int _, int __)
    cdef int _sta(self, int addr, int _)
    cdef int _stx(self, int addr, int _)
    cdef int _sty(self, int addr, int _)
    cdef int _tax(self, int _, int __)
    cdef int _tay(self, int _, int __)
    cdef int _tsx(self, int _, int __)
    cdef int _txa(self, int _, int __)
    cdef int _txs(self, int _, int __)
    cdef int _tya(self, int _, int __)



def load_rom_raw(filename):
    """
    Loads the data from a .nes file as the nesheader and the rom data without further processing
    :param filename:
    :return:
    """
    with open(filename, "rb") as f:
        nesheader = f.read(16)
        bytecode = f.read()
    return bytecode, nesheader

def load_palette(filename):
    """
    Loads a 64 value palette from a .pal file (binary, 3 bytes per color).
    Palettes can be created from bisqwit's tool here: https://bisqwit.iki.fi/utils/nespalette.php
    :return: A list of rgb tuples (elements in range 0-255) corresponding to the palette colors
    """
    with open(filename, "rb") as f:
        pal = f.read()
    palette = []
    for i in range(64):
        palette.append( (pal[i * 3], pal[i * 3 + 1], pal[i * 3 + 2]) )
    return palette



from .memory cimport NESVRAM
from .system cimport InterruptListener

### NES PPU Constants ##################################################################################################

# this odd mechanism allows (integer) constants to be shared between pyx files via pxd files
# if these constants also need to be used from python use cpdef in place of cdef here
cdef enum:
    NUM_PPU_REGISTERS = 8

    # Register indices (this is not just an enum, this is the offset of the register in the CPU memory map from 0x2000)
    PPU_CTRL = 0
    PPU_MASK = 1
    PPU_STATUS = 2
    OAM_ADDR = 3
    OAM_DATA = 4
    PPU_SCROLL = 5
    PPU_ADDR = 6
    PPU_DATA = 7

    # masks for the bits in ppu registers
    # ppu_status
    VBLANK_MASK =               0b10000000  # same for ppu_ctrl
    SPRITE0_HIT_MASK =          0b01000000  #
    SPRITE_OVERFLOW_MASK =      0b00100000  #

    # ppu_ctrl
    SPRITE_SIZE_MASK =          0b00100000  #
    BKG_PATTERN_TABLE_MASK =    0b00010000  #
    SPRITE_PATTERN_TABLE_MASK = 0b00001000  #
    VRAM_INCREMENT_MASK =       0b00000100  #
    #NAMETABLE_MASK =            0b00000011

    BIT_NAMETABLE_X = 0
    BIT_NAMETABLE_Y = 1

    # ppu_mask
    RENDERING_ENABLED_MASK =    0b00011000
    RENDER_SPRITES_MASK =       0b00010000
    RENDER_BACKGROUND_MASK =    0b00001000
    #GREYSCALE_MASK =            0b00000001

     # bit numbers of some important bits in registers
    # ppu_status
    V_BLANK_BIT = 7             # same for ppu_ctrl

    # ppu mask
    RENDER_LEFT8_BKG_BIT     = 1
    RENDER_LEFT8_SPRITES_BIT = 2

    # byte indices in ppu scroll
    PPU_SCROLL_X = 0
    PPU_SCROLL_Y = 1

    # size of the OAM memory
    OAM_SIZE_BYTES = 256

    # screen and sprite/tile sizes:
    PIXELS_PER_LINE = 341       # number of pixels per ppu scanline; only 256 of thes are visible
    SCREEN_WIDTH_PX = 256       # visible screen width (number of visible pixels per row)
    SCREEN_HEIGHT_PX = 240      # visible screen height (number of visible rows)
    VERTICAL_OVERSCAN_PX = 8    # The NES assumes that the top and bottom 8 rows will not be visible due to CRT overscan
    HORIZONTAL_OVERSCAN_PX = 8  # Some games look better with horizontal overscan s
                                # see https://wiki.nesdev.com/w/index.php/Overscan
    PRERENDER_LINE = 261        # prerender scanline

    SCREEN_TILE_ROWS = 30
    SCREEN_TILE_COLS = 32
    TILE_HEIGHT_PX = 8
    TILE_WIDTH_PX = 8

    # size of a tile in the pattern table in bytes (width_px(8) x height_px(8) x bits_per_px(2) / bits_per_byte(8) == 16)
    PATTERN_SIZE_BYTES = 16


### NES PPU prototype ##################################################################################################
cdef class NESPPU:
    cdef:
        # ppu registers
        unsigned char ppu_ctrl, ppu_mask, oam_addr, oam_data, _ppu_data_buffer, _io_latch
        unsigned char ppu_scroll[2]
        unsigned int ppu_addr, _ppu_byte_latch

        # ppu current position
        int line, pixel  # screen drawing starts at pixel 1 (not 0) and runs to pixel 256.

        # status flags
        bint in_vblank, sprite_zero_hit, sprite_overflow, ignore_ppu_ctrl

        # internal tracking counters
        int frames_since_reset, time_at_new_frame
        long long cycles_since_reset

        # Sprite rendering
        unsigned char oam[OAM_SIZE_BYTES]      # main internal OAM memory
        unsigned char _oam[32]                 # a secondary internal OAM array to store sprites active on next scanline
        int _active_sprite_addrs[8]            # addresses of the active sprites
        unsigned char _sprite_bkg_priority[8]  # a buffer to hold whether a given sprite has priority over background
        int _sprite_line[8]                    # lines of the active sprite that we are on
        char _sprite_pattern[8][8]             # decoded patterns for the active sprites
        int _num_active_sprites                # how many sprites are active in this current line
        bint irq_tick_triggers[68]             # whether or not an irq tick is triggered on this pixel of sprite fetch

        # Background rendering
        unsigned int _pattern_lo, _pattern_hi   # 16-bit bkg pattern registers (only bottom 16 bits relevant)
        int _effective_x, _effective_y          # current background tile position being fetched
        int _palette[2][4]                      # palettes for next tiles

        # access to other bits of the system - vram and interrupts
        NESVRAM vram
        InterruptListener interrupt_listener

        # the screen buffer itself; currently uses a packed 32 bit int with pixel format xRGB
        unsigned int screen_buffer[256][240]

        # palettes for all the colors the NES can display
        #unsigned int rgb_palette[64][3]    # in standard RGB format
        unsigned int hex_palette[64]       # in packed 32 bit xRGB format

        # special colors that are in use
        int transparent_color, bkg_color

        # caches for palettes, since decoding them is a bit slow
        int _palette_cache[8][4]
        int _palette_cache_valid[8]

    ########### functions ##############################################

    # write OAM data (called from OAM DMA)
    cdef void write_oam(self, unsigned char* data)

    # screen buffer copy and clear
    cdef int get_background_color(self)
    cpdef void copy_screen_buffer_to(self, unsigned int[:, :] dest, bint v_overscan=?, bint h_overscan=?)
    cdef void _clear_to_bkg(self)

    # registers read/write
    cdef unsigned char read_register(self, int register)
    cdef void write_register(self, int register, unsigned char value)
    cdef unsigned char ppu_status(self)

    # running ppu cycles
    cdef int run_cycles(self, int num_cycles)
    cdef void prerender_scanline(self)
    cdef void render_visible_scanline(self)
    cdef void increment_pixel(self)
    cdef void _increment_vram_address(self)
    cdef void _trigger_nmi(self)
    cdef void _reset_effective_y(self)

    # sprite rendering
    cdef void _prefetch_active_sprites(self)
    cdef void _fill_sprite_latches(self, int double_sprites)
    cdef int _overlay_sprites(self, int bkg_pixel)

    # background rendering
    cdef void fill_bkg_latches(self)
    cdef int _get_bkg_pixel(self)

    # palette decoding and caching
    cdef void decode_palette(self, int* palette_out, int palette_id, bint is_sprite)
    cdef void invalidate_palette_cache(self)

    # debug
    cpdef void debug_render_nametables(self, unsigned int[:, :] dest)
    cpdef void debug_render_tile(self, unsigned int[:, :] dest, int x0, int y0, int tile_index, int tile_bank, int palette_id, bint flip_h, bint flip_v)





#import pyximport; pyximport.install()

from .bitwise cimport bit_high
try:
    import pyaudio
    has_audio = False
except ImportError:
    has_audio = True

from .system cimport DMC_DMA

#### Length Table and other constant arrays ############################################################################

cdef int[32] LENGTH_TABLE = [ 10, 254, 20,  2, 40,  4, 80,  6, 160,  8, 60, 10, 14, 12, 26, 14,
                              12,  16, 24, 18, 48, 20, 96, 22, 192, 24, 72, 26, 16, 28, 32, 30 ]

#### Basic components ##################################################################################################

cdef class APUUnit:
    """
    Base class for the APU's sound generation units providing some basic common functionality
    """
    def __init__(self):
        self.enable = False
        self.length_ctr = 0
        self.ctr_halt = False
        for i in range(SAMPLE_RATE):
            self.output[i] = 0

    cdef void update_length_ctr(self):
        if not self.ctr_halt:
            self.length_ctr = (self.length_ctr - 1) if self.length_ctr > 0 else 0

    cdef void set_enable(self, bint value):
        self.enable = value
        if not self.enable:
            self.length_ctr = 0

    cdef void set_length_ctr(self, int value):
        if self.enable:
            self.length_ctr = LENGTH_TABLE[value & 0b11111]


cdef class APUEnvelope:
    """
    Volume envelope unit used in pulse and noise APU units
    Reference:
        [6] envelope:  https://wiki.nesdev.com/w/index.php/APU_Envelope
    """
    def __init__(self):
        self.start_flag = False
        self.loop_flag = False
        self.decay_level = 0
        self.divider = 0
        self.volume = 0

    cdef void restart(self):
        self.start_flag = False
        self.decay_level = 15
        self.divider = self.volume

    cdef void update(self):
        if not self.start_flag:   # if start flag is clear
            # clock divider
            if self.divider == 0:
                # When divider is clocked while at 0, it is loaded with volume and clocks the decay level counter [6]
                self.divider = self.volume
                # clock decay counter:
                #   if the counter is non-zero, it is decremented, otherwise if the loop flag is set, the decay level
                #   counter is loaded with 15. [6]
                if self.decay_level > 0:
                    self.decay_level -=1
                elif self.loop_flag:
                    self.decay_level = 15
            else:
                # clock divider (is this right?)
                self.divider -= 1
        else:
            self.restart()


#### Sound generation units ############################################################################################

cdef class APUTriangle(APUUnit):
    """
    APU unit for generating triangle waveform
    Reference:
        [4] triangle:  https://wiki.nesdev.com/w/index.php/APU_Triangle
    """
    def __init__(self):
        super().__init__()
        self.period = 0
        self.phase = 0

        self.linear_reload_flag = False
        self.linear_reload_value = 0
        self.linear_ctr = 0

    cdef void write_register(self, int address, unsigned char value):
        """
        Set properties of the triangle waveform generator from a write to its registers.
        """
        if address == 0x4008:
            self.ctr_halt = bit_high(value, 7)
            # don't set the counter directly, just set the reload value for now
            self.linear_reload_value = value & 0b01111111
        elif address == 0x400A:
            self.period = (self.period & 0xFF00) + value
        elif address == 0x400B:
            self.period = (self.period & 0xFF) + ((value & 0b00000111) << 8)
            self.set_length_ctr(value >> 3)
            self.linear_reload_flag = True

    cdef void quarter_frame(self):

        # Update triangle linear counter.  This is a bit complicated and occurs as follows [4]:
        # if counter reload flag is set:
        #     linear counter <-- counter reload value
        # elif linear counter > 0:
        #     decrement linear counter
        # if control flag clear:
        #     counter reload flag cleared
        if self.linear_reload_flag:
            self.linear_ctr = self.linear_reload_value
        elif self.linear_ctr > 0:
             self.linear_ctr -= 1

        if not self.ctr_halt:  # this is also the control flag
            self.linear_reload_flag = False

    cdef void half_frame(self):
        self.update_length_ctr()

    cdef int generate_sample(self):
        """
        Generate a single sample of the triangle wave and advance the phase appropriately
        """
        cdef double freq_hz
        cdef int v

        # frequency of the triangle wave is given by timer as follows [4]:
        freq_hz = CPU_FREQ_HZ * 1. / (32. * (self.period + 1))

        # how much phase we step with each cycle is given by
        #   phase_per_samp = cycles per sample = (freq_hz / samples per second)
        # unit of phase here is CYCLES not radians (i.e. 1 cycle = 2pi radians)
        phase_per_samp = freq_hz / SAMPLE_RATE

        # if the triangle wave is not enabled or its linear or length counter is zero, this is zero.
        # Also added here is an exclusion for ultrasonic frequencies, which is used in MegaMan to silence the triangle
        # this is not entirely accurate, but probably produces nicer sounds.
        if (not self.enable) or self.length_ctr == 0 or self.linear_ctr == 0 or self.period < 2:
            v = 0
        else:
            v = int(31.999999 * abs(0.5 - self.phase))    # int cast rounds down, should never be 16
            self.phase = (self.phase + phase_per_samp) % 1.

        return v


cdef class APUPulse(APUUnit):
    """
    APU pulse unit; there are two of these in the APU
    Reference:
        [5] pulse: https://wiki.nesdev.com/w/index.php/APU_Pulse
        [12] http://nesdev.com/apu_ref.txt
        [13] https://wiki.nesdev.com/w/index.php/APU_Sweep
    """
    # Duty cycles for the pulse generators [5]
    DUTY_CYCLES = [[0, 1, 0, 0,  0, 0, 0, 0],
                   [0, 1, 1, 0,  0, 0, 0, 0],
                   [0, 1, 1, 1,  1, 0, 0, 0],
                   [1, 0, 0, 1,  1, 1, 1, 1]]

    def __init__(self, is_unit_1):
        super().__init__()
        self.constant_volume = False
        self.period = 1
        self.adjusted_period = 1   # period adjusted by the sweep units
        self.duty = 0
        self.phase = 0
        self.env = APUEnvelope()
        self.is_unit_1 = is_unit_1

        self.sweep_enable = False
        self.sweep_negate = False
        self.sweep_reload = False
        self.sweep_period = 1
        self.sweep_shift = 0
        self.sweep_divider = 0

        # copy the pulse duty cycle patterns into an int array
        # todo: is there a way to have shared cython class-level variables or constants?
        for i in range(4):
            for j in range(8):
                self.duty_waveform[i][j] = self.DUTY_CYCLES[i][j]

    cdef void write_register(self, int address, unsigned char value):
        """
        Set properties of the pulse waveform generators from a write to one of their registers; assumes address has been
        mapped to the 0x4000-0x4003 range (i.e. if this is pulse 1, subtract 4)
        """
        address -= 4 if not self.is_unit_1 else 0
        if address == 0x4000:
            self.duty = (value & 0b11000000) >> 6
            self.ctr_halt = bit_high(value, 5)         # } these are the same
            self.env.loop_flag = bit_high(value, 5)    # }
            self.constant_volume = bit_high(value, 4)
            self.env.volume = value & 0b00001111
        elif address == 0x4001:
            self.sweep_enable = bit_high(value, 7)
            self.sweep_period = ((value & 0b01110000) >> 4) + 1
            self.sweep_negate = bit_high(value, 3)
            self.sweep_shift = value & 0b00000111
            self.sweep_reload = True
            #print("sweep: {:08b}".format(value))
        elif address == 0x4002:
            # timer low
            self.period = (self.period & 0xFF00) + value
            self.adjusted_period = self.period
        elif address == 0x4003:
            self.period = (self.period & 0xFF) + ((value & 0b00000111) << 8)
            self.adjusted_period = self.period
            self.set_length_ctr(value >> 3)
            # side effect: the sequencer is restarted at the first value of the sequence and envelope is restarted [5]
            self.phase = 0
            # side effect: restart envelope and set the envelope start flag
            self.env.restart()
            self.env.start_flag = True

    cdef void quarter_frame(self):
        self.env.update()

    cdef void half_frame(self):
        self.sweep_update()
        self.update_length_ctr()

    cdef void sweep_update(self):
        """
        Adjust the period based on the sweep unit.  Part of the functionality here is based on the following paragraph
        from [12]:
            "When the channel's period is less than 8 or the result of the shifter is
             greater than $7FF, the channel's DAC receives 0 and the sweep unit doesn't
             change the channel's period. Otherwise, if the sweep unit is enabled and the
             shift count is greater than 0, when the divider outputs a clock, the channel's
             period in the third and fourth registers are updated with the result of the
             shifter."
        This seems to disagree a little with [13] in that the description in [13] permits zero as a shift
        value (and even uses it as an example), whereas the above text excludes it.  This makes a difference in the
        game Ghengis Khan, which if 0 is permitted as a shift has background music which sounds wrong (and differs from
        FCEUX) in the main game screens.

        In the system here, adjusted_period is used to determine whether or not the period has been changed outside of
        the permitted range, silencing the channel (without adjusting period outside that range).  Don't know if this is
        correct, but it seems to be what is implied in [12].

        Finally, should the change amount be based on the non-adjusted period or the adjusted period?  I.e. should the
        period be irrevocably altered by the sweep unit, or is the original period retained somewhere and used to
        calculate the shift each time?  In a recurring multiple shift, this changes the frequency change rate from
        exponential to linear.  No clear idea of what is correct.
        """
        cdef int change_amount
        if self.sweep_reload:
            # If [the divider's counter is zero or] the reload flag is true, the counter is set to P and the reload flag
            # is cleared. [7]
            self.sweep_divider = self.sweep_period
            self.sweep_reload = False
            return

        if self.sweep_divider > 0:
            # Otherwise, the counter is decremented. [7]
            self.sweep_divider -= 1
        else: # self.sweep_divider == 0:
            # If the divider's counter is zero [...], the counter is set to P and the reload flag is cleared. [7]
            self.sweep_divider = self.sweep_period
            self.sweep_reload = False

            if self.sweep_enable and self.sweep_shift > 0:
                # in this case, trigger a period adjustment
                change_amount = self.period >> self.sweep_shift
                if self.sweep_negate:
                    change_amount = -change_amount
                    if not self.is_unit_1:
                        change_amount -= 1

                # check if the adjusted period would go outside range; if so, don't update main period, but channel
                # will be silenced (see description above)
                self.adjusted_period = self.period + change_amount
                if 8 < self.adjusted_period <= 0x7FF:
                    self.period += change_amount

    cdef int generate_sample(self):
        """
        Generate one output sample from the pulse unit.
        """
        cdef double freq_hz
        cdef int v, volume, change_amount

        freq_hz = CPU_FREQ_HZ * 1. / (16. * (self.period + 1))

        # how much phase we step with each cycle is given by
        #   phase_per_samp = cycles per sample = (freq_hz / samples per second)
        # unit of phase here is CYCLES not radians (i.e. 1 cycle = 2pi radians)
        phase_per_samp = freq_hz / SAMPLE_RATE

        # there are several conditions under which the channel is muted.  [7] and others.
        if ( not self.enable
             or self.length_ctr == 0
             or self.adjusted_period < 8
             or self.adjusted_period > 0x7FF
            ):
            v = 0
        else:
            volume = self.env.volume if self.constant_volume else self.env.decay_level
            v = volume * self.duty_waveform[self.duty][int(7.999999 * self.phase)]
            self.phase = (self.phase + phase_per_samp) % 1.
        return v


cdef class APUNoise(APUUnit):
    """
    APU pulse unit; there are two of these in the APU
    Reference:
        [8] noise: https://wiki.nesdev.com/w/index.php/APU_Noise

    """
    TIMER_TABLE = [4, 8, 16, 32, 64, 96, 128, 160, 202, 254, 380, 508, 762, 1016, 2034, 4068]

    def __init__(self):
        super().__init__()
        self.constant_volume = False
        self.mode = False
        self.period = 4
        self.feedback = 1  # loaded with 1 on power up [8]
        self.timer = 0
        self.env = APUEnvelope()
        for i in range(16):
            self.timer_table[i] = self.TIMER_TABLE[i]

    cdef void write_register(self, int address, unsigned char value):
        """
        Set properties of the noise waveform generators from a write to its registers.
        """
        if address == 0x400C:
            self.ctr_halt = bit_high(value, 5)    # } these are the same
            self.env.loop_flag = bit_high(value, 5)     # }
            self.constant_volume = bit_high(value, 4)
            self.env.volume = value & 0b00001111
        elif address == 0x400E:
            self.mode = bit_high(value, 7)
            self.period = self.timer_table[value & 0b00001111]
        elif address == 0x400F:
            self.set_length_ctr(value >> 3)
            # side effect: restart envelope and set the envelope start flag
            self.env.restart()
            self.env.start_flag = True

    cdef void quarter_frame(self):
        self.env.update()

    cdef void half_frame(self):
        self.update_length_ctr()

    cdef void update_cycles(self, int cycles):
        cdef int xor_bit, feedback_bit
        self.timer += cycles

        if self.timer >= 2 * self.period:
            self.timer -= 2 * self.period
            xor_bit = 6 if self.mode else 1
            feedback_bit = bit_high(self.feedback, 0) ^ bit_high(self.feedback, xor_bit)
            self.feedback >>= 1
            self.feedback |= (feedback_bit << 14)

    cdef int generate_sample(self):
        """
        Generates a noise sample.  Updates a shift register to create pseudo-random samples and applies the
        noise volume envelope.
        """
        # clock the feedback register 0.5 * CPU_FREQ / SAMPLE_RATE / noise_period times per sample
        volume = self.env.volume if self.constant_volume else self.env.decay_level
        if not self.enable or self.length_ctr==0:
            return 0
        return volume * (self.feedback & 1)   # bit here should actually be negated, but doesn't matter


cdef class APUDMC(APUUnit):
    """
    The delta modulation channel (DMC) of the APU.  This allows delta-coded 1-bit samples to be played
    directly from memory.  It operates in a different way to the other channels, so is not based on them.
    References:
        [9] DMC: https://wiki.nesdev.com/w/index.php/APU_DMC
        [10] DMC interrupt flag reset: http://www.slack.net/~ant/nes-emu/apu_ref.txt
        [11] interesting DMC model: http://www.slack.net/~ant/nes-emu/dmc/
    """
    RATE_TABLE = [428, 380, 340, 320, 286, 254, 226, 214, 190, 160, 142, 128, 106,  84,  72,  54]

    def __init__(self, interrupt_listener):
        super().__init__()

        self.interrupt_listener = interrupt_listener

        self.enable = False
        self.silence = True
        self.irq_enable = False
        self.loop_flag = False
        self.interrupt_flag = False

        self.rate = 0
        self.output_level = 0
        self.sample_address = 0
        self.address = 0
        self.sample_length = 0
        self.bytes_remaining = 0
        self.sample = 0
        self.timer = 0

        # copy the rate table lookup into the cdef'ed variable for speed
        for i in range(16):
            self.rate_table[i] = self.RATE_TABLE[i]

    cdef void write_register(self, int address, unsigned char value):
        """
        Write a value to one of the dmc's control registers; called from APU write register.  Address must be in the
        0x4010 - 0x4013 range (inclusive); if not the write is ignored
        """
        if address == 0x4010:
            self.irq_enable = bit_high(value, 7)
            if not self.irq_enable:
                self.interrupt_flag = False
            self.loop_flag = bit_high(value, 6)
            self.rate = self.rate_table[value & 0b00001111]
        elif address == 0x4011:
            self.output_level = value & 0b01111111
        elif address == 0x4012:
            self.sample_address = 0xC000 + (value << 6)
            self.address = self.sample_address
        elif address == 0x4013:
            self.sample_length = (value << 4) + 1
            self.bytes_remaining = self.sample_length

    cdef void update_cycles(self, int cpu_cycles):
        """
        Update that occurs on every CPU clock tick, run cpu_cycles times
        """
        cdef int v
        self.timer += cpu_cycles

        if self.timer < 2 * self.rate:
            return

        # now the unit's timer has ticked, so update the unit
        self.timer -= 2 * self.rate

        # update bits_remaining counter
        if self.bits_remaining == 0:
            # cycle end; a new cycle can start
            self.bits_remaining = 8
            self.read_advance()

        # read the bit
        v = self.sample & 1

        if not self.silence:
            if v == 0 and self.output_level >= 2:
                self.output_level -= 2
            elif v == 1 and self.output_level <= 125:
                self.output_level += 2

        # clock the shift register one place to the right
        self.sample >>= 1
        self.bits_remaining -= 1

    cdef void read_advance(self):
        """
        Reads a byte of memory and places it into the sample buffer; advances memory pointer, wrapping if necessary.
        """
        if self.bytes_remaining == 0:
            if self.loop_flag:
                # if looping and have run out of data, go back to the start
                self.bytes_remaining = self.sample_length
                self.address = self.sample_address
            else:
                self.silence = True

        if self.bytes_remaining > 0:
            self.sample = self.memory.read(self.address)
            self.address = (self.address + 1) & 0xFFFF
            self.bytes_remaining -= 1
            self.silence = False
            if self.bytes_remaining == 0 and self.irq_enable:
                # this was the last byte that we just read
                # "the IRQ is generated when the last byte of the sample is read, not when the last sample of the
                # sample plays" [9]
                self.interrupt_listener.raise_irq()
                self.interrupt_flag = True

            # a DMC memory read should stall the CPU here for a variable number of cycles
            self.interrupt_listener.raise_dma_pause(DMC_DMA)

    cdef int generate_sample(self):
        """
        Generate the next DMC sample.
        """
        return self.output_level


#### The APU ###########################################################################################################

cdef class NESAPU:
    """
    NES APU

    Sources:
        [1] https://wiki.nesdev.com/w/index.php/APU#Registers
        [2] https://wiki.nesdev.com/w/index.php/APU_Frame_Counter
        [3] https://wiki.nesdev.com/w/index.php/APU_Length_Counter

        [10] DMC interrupt flag reset: http://www.slack.net/~ant/nes-emu/apu_ref.txt
    """
    def __init__(self, interrupt_listener, master_volume=0.5):
        self.interrupt_listener = interrupt_listener

        # Power-up and reset have the effect of writing $00 to apu status (0x4015), silencing all channels [1]
        self.frame_segment = 0
        self.cycles = 0
        self._reset_timer_in = -1
        self.samples_per_cycle = SAMPLE_RATE * 1. / CPU_FREQ_HZ
        self.samples_required = 0
        self.rate=SAMPLE_RATE
        self.irq_inhibit = True

        # the frame_interrupt_flag is connected to the CPU's IRQ line, so changes to this flag should be accompanied by
        # a change to the interrupt_listener's state
        self.frame_interrupt_flag = False

        # sound output buffer (this is a ring buffer, so these variables track the current start and end position)
        self._buffer_start = 0
        self._buffer_end = 1600  # give it some bonus sound to start with

        self.master_volume = master_volume
        self.mode = FOUR_STEP

        # sound production units
        self.triangle = APUTriangle()
        self.pulse1 = APUPulse(is_unit_1=True)
        self.pulse2 = APUPulse(is_unit_1=False)
        self.noise = APUNoise()
        self.dmc = APUDMC(interrupt_listener)

        for i in range(APU_BUFFER_LENGTH):
            self.output[i] = 0

    cpdef short[:] get_sound(self, int samples):
        """
        Generate samples of audio using the current audio settings.  The number of samples generated
        should be small (probably at most about 1/4 frame - 200 samples at 48kHz - to allow all effects to be
        reproduced, however, somewhat longer windows can probably be used; there are 800 samples in a frame at 48kHz).
        The absolute maximum that will be returned is 1s of audio.
        """
        cdef int i
        samples = min(samples, CHUNK_SIZE, self._buffer_end - self._buffer_start)
        for i in range(samples):
            self.buffer[i] = self.output[(self._buffer_start + i) & (APU_BUFFER_LENGTH - 1)]
        self._buffer_start += samples
        cdef short[:] data = <short[:samples]>self.buffer
        return data

    cdef void generate_sample(self):
        tri = self.triangle.generate_sample()
        p1 = self.pulse1.generate_sample()
        p2 = self.pulse2.generate_sample()
        noise = self.noise.generate_sample()
        dmc = self.dmc.generate_sample()

        v = self.mix(tri, p1, p2, noise, dmc)

        self.output[self._buffer_end & (APU_BUFFER_LENGTH - 1)] = v
        self._buffer_end += 1

    cpdef int buffer_remaining(self):
        return self._buffer_end - self._buffer_start

    cpdef void set_volume(self, float volume):
        self.master_volume = volume

    cpdef void set_rate(self, int rate):
        self.rate = rate
        self.samples_per_cycle = self.rate * 1. / CPU_FREQ_HZ

    cpdef int get_rate(self):
        return self.rate

    ######## interfacing with pyaudio #####################
    # keep all pyaudio code in this section

    def pyaudio_callback(self, in_data, frame_count, time_info, status):
        if self.buffer_remaining() > 0:
            data = self.get_sound(frame_count)
            return (data, pyaudio.paContinue)
        else:
            return (None, pyaudio.paAbort)

    ########################################################

    cdef int run_cycles(self, int cpu_cycles):
        """
        Updates the APU by the given number of cpu cycles.  This updates the frame counter if
        necessary (every quarter or fifth video frame).  Timings from [2].
        """
        cdef int new_segment, cpu_cycles_per_loop, cycles
        cdef bint quarter_frame = False, force_ticks = False

        while cpu_cycles > 0:
            cycles = cpu_cycles if cpu_cycles < MAX_CPU_CYCLES_PER_LOOP else MAX_CPU_CYCLES_PER_LOOP
            self.cycles += cycles
            cpu_cycles -= MAX_CPU_CYCLES_PER_LOOP

            self.dmc.update_cycles(cycles)
            self.noise.update_cycles(cycles)

            self.samples_required += cycles * self.samples_per_cycle
            while self.samples_required > 1:
                self.generate_sample()
                self.samples_required -= 1

            if self._reset_timer_in >= 0:
                self._reset_timer_in -= cycles
                if self._reset_timer_in < 0:
                    self.cycles = 0
                    if self.mode == FIVE_STEP:
                        force_ticks = True
                    else: # four step mode
                        # If mode is FOUR_STEP, do *not* generate frame ticks
                        self.frame_segment = 0

            if self.cycles < 7457:
                new_segment = 0
            elif self.cycles < 14913:
                new_segment = 1
            elif self.cycles < 22371:
                new_segment = 2
            elif self.cycles <= 29829:   # this should be <, but logic is easier this way
                new_segment = 3
            else:
                if self.mode == FOUR_STEP:
                    new_segment = 0
                    self.cycles -= 29830
                    if not self.irq_inhibit:
                        self.interrupt_listener.raise_irq()
                        self.frame_interrupt_flag = True
                else:  # five-step counter
                    if self.cycles <= 37281:
                        new_segment = 5
                    else:
                        new_segment = 0
                        self.cycles -= 37282

            if self.frame_segment != new_segment or force_ticks:
                if self.mode == FOUR_STEP or new_segment != 3:
                    # the quarter frame tick happens on the 0, 1, 2, 4 ticks in FIVE_STEP mode
                    # source: (https://wiki.nesdev.com/w/index.php/APU) section on Frame Counter
                    self.quarter_frame_tick()
                quarter_frame = True
                if new_segment == 0 or new_segment == 2 or force_ticks:
                    self.half_frame_tick()

            self.frame_segment = new_segment
        return quarter_frame

    cdef void quarter_frame_tick(self):
        """
        This is a tick that happens four times every (video) frame.  It updates the envelopes and the
        linear counter of the triange generator [2].
        """
        self.triangle.quarter_frame()
        self.pulse1.quarter_frame()
        self.pulse2.quarter_frame()
        self.noise.quarter_frame()

    cdef void half_frame_tick(self):
        """
        This is a tick that happens twice every (video) frame.  It updates the length counters and the
        sweep units [2].
        """
        self.triangle.half_frame()
        self.pulse1.half_frame()
        self.pulse2.half_frame()
        self.noise.half_frame()

    cdef unsigned char read_register(self, int address):
        """
        Read an APU register.  Actually the only one you can read is STATUS (0x4015).
        """
        cdef unsigned char value
        cdef bint dmc_active = False

        dmc_active = self.dmc.bytes_remaining > 0

        if address == STATUS:
            value = (  (self.dmc.interrupt_flag << 7)
                     + (self.frame_interrupt_flag << 6)
                     + (dmc_active << 4)
                     + ((self.noise.length_ctr > 0) << 3)
                     + ((self.triangle.length_ctr > 0) << 2)
                     + ((self.pulse2.length_ctr > 0) << 1)
                     + (self.pulse1.length_ctr > 0)
                    )
            self.frame_interrupt_flag = False
            # "When $4015 is written to, the channels' length counter enable flags are set,
            # the DMC is possibly started or stopped, and the DMC's IRQ occurred flag is cleared." [10]
            self.dmc.interrupt_flag = False
            self.interrupt_listener.reset_irq()
            return value

        print("apu read: {:04X}".format(address))

    cdef void write_register(self, int address, unsigned char value):
        """
        Write to one of the APU registers.
        """
        cdef APUPulse pulse

        if address == STATUS:
            self._set_status(value)
        elif address == FRAME_COUNTER:
            self.mode = bit_high(value, BIT_MODE)
            self.irq_inhibit = bit_high(value, BIT_IRQ_INHIBIT)
            if self.irq_inhibit:
                self.frame_interrupt_flag = False
                self.interrupt_listener.reset_irq()
            # side effects:  reset timer (in 3-4 cpu cycles' time, if mode set generate quarter and half frame signals)
            self._reset_timer_in = 3 + self.cycles % 2
        elif 0x4000 <= address <= 0x4007:
            # a pulse register
            # there are two pulse registers set by 0x4000-0x4003 and 0x4004-0x4007; select the correct one to update
            pulse = self.pulse1 if address < 0x4004 else self.pulse2
            pulse.write_register(address, value)
        elif 0x4008 <= address <= 0x400B:
            # a triangle register
            self.triangle.write_register(address, value)
        elif 0x400C <= address <= 0x400F:
            self.noise.write_register(address, value)
        elif 0x4010 <= address <= 0x4013:
            self.dmc.write_register(address, value)

    cdef void _set_status(self, unsigned char value):
        """
        Sets up the status register from a write to the status register 0x4015
        """
        self.triangle.set_enable(bit_high(value, BIT_ENABLE_TRIANGLE))
        self.pulse1.set_enable(bit_high(value, BIT_ENABLE_PULSE1))
        self.pulse2.set_enable(bit_high(value, BIT_ENABLE_PULSE2))
        self.noise.set_enable(bit_high(value, BIT_ENABLE_NOISE))
        self.dmc.set_enable(bit_high(value, BIT_ENABLE_DMC))

    cdef int mix(self, int triangle, int pulse1, int pulse2, int noise, int dmc):
        """
        Mix the channels into signed 16-bit audio samples
        """
        cdef double pulse_out, tnd_out, sum_pulse, sum_tnd

        sum_pulse = pulse1 + pulse2
        sum_tnd = (triangle / 8227.) + (noise / 12241.) + (dmc / 22638.)
        pulse_out = 95.88 / ((8128. / sum_pulse) + 100.) if sum_pulse != 0 else 0
        tnd_out = 159.79 / (1. / sum_tnd + 100.) if sum_tnd != 0 else 0
        return int( ((pulse_out + tnd_out) ) * self.master_volume * SAMPLE_SCALE)



from .memory import MemoryBase

class CartBase(MemoryBase):
    """
    Cartridges can provide memory (with separate address spaces) to the CPU and the PPU.
    To accommodate this, additional read_ppu and write_ppu methods provide access to the PPU-linked memory.
    """

    def __init__(self):
        super().__init__()
        self.nametable_mirror_pattern=None

    def read_ppu(self, address):
        raise NotImplementedError()

    def write_ppu(self, address, v):
        raise NotImplementedError()



class NESCart0(CartBase):
    """
    Basic NES Cartridge (Type 0 / MMC0).  Consists of up to 8kB RAM, 32kB PRG ROM, 8kB CHR ROM
    """
    # CPU memory space
    RAM_START = 0x6000
    PRG_ROM_START = 0x8000

    # PPU memory space
    CHR_ROM_START = 0x0000

    def __init__(self, prg_rom_data=None, chr_rom_data=None, ram_size_kb=8, prg_start_addr=None, nametable_mirror_pattern=(0,0,1,1)):
        super().__init__()

        # initialize ram (CPU connected)
        if ram_size_kb not in [2, 4, 8]:
            raise ValueError("Cart 0 ram size should be 2, 4 or 8kB")
        self.ram = bytearray(ram_size_kb * 1024)

        # initialize prg rom from supplied data  (CPU connected)
        self.prg_rom = bytearray(prg_rom_data)

        if len(self.prg_rom) not in [16 * 1024, 32 * 1024]:
            raise ValueError("Cart 0 prg rom size should be 16 or 32kB")

        # initialize chr rom from supplied data  (PPU connected)
        # or create ram if there isn't one
        if chr_rom_data:
            self.chr_mem = bytearray(chr_rom_data)
            if len(self.chr_mem) != 8 * 1024:
                raise ValueError("Cart 0 chr rom size should be 8kB")
        else:
            self.chr_mem = bytearray(8 * 1024)   # make a RAM

        self.prg_start_addr = prg_start_addr if prg_start_addr else self.PRG_ROM_START
        self.nametable_mirror_pattern = nametable_mirror_pattern
        #rom_data_start = load_rom_at - self.PRG_ROM_START
        #self.rom[rom_data_start:rom_data_start + len(rom_data)] = rom_data

    def read(self, address):
        if address < self.PRG_ROM_START:
            # ram access
            return self.ram[address % len(self.ram)]
        else:
            return self.prg_rom[(address - self.prg_start_addr) % len(self.prg_rom)]

    def write(self, address, value):
        if address < self.PRG_ROM_START:
            # ram access
            self.ram[address % len(self.ram)] = value
        else:
            # should not be able to write to ROM, write here has no effect
            print("WARNING: OVERWRITING PRG ROM")
            self.prg_rom[(address - self.prg_start_addr) % len(self.prg_rom)] = value

    def read_ppu(self, address):
        return self.chr_mem[address % len(self.chr_mem)]

    def write_ppu(self, address, value):
        print("WARNING: OVERWRITING CHR ROM")
        self.chr_mem[address % len(self.chr_mem)] = value
